# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
class AppyError(Exception):
    '''Root Appy exception class'''

class ValidationError(AppyError):
    '''Represents an error that occurs on data sent to the Appy server'''

class InternalError(AppyError):
    '''Represents a programming error: something that should never occur'''

class CommercialError(AppyError):
    '''Raised when some functionality is called from the commercial version but
       is available only in the free, open source version.'''
    MSG = 'This feature is not available in the commercial version. It is ' \
          'only available in the free, open source (GPL) version of Appy.'
    def __init__(self): AppyError.__init__(self, self.MSG)
# ------------------------------------------------------------------------------



'''Execute commands to a Subversion repository'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import os
from pathlib import Path

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Subversion:
    '''Represents a distant Subversion server'''

    # svn options for non interactive mode
    nonInteractive = '--non-interactive --no-auth-cache'

    def __init__(self, specifier):
        # A p_specifier is an object with attributes such as "url", "name",
        # "login", "password"...
        self.specifier = specifier

    def build(self, command, path):
        '''Builds the Subversion p_command based on info from p_self.specifier
           and the concerned local p_path.'''
        # Returns a tuple (v_command, v_path), v_command being the complete
        # Subversion command (p_command and added args), and v_path being the
        # p_path completed with the local folder name in <site>/lib.
        r = ['svn', command]
        spec = self.specifier
        # Compute the local folder name in <site>/lib
        if spec.name:
            local = spec.name
        else:
            # The name of the app to checkout is the last part of spec.url
            local = Path(spec.url).name
        local = path / local        
        if command == 'checkout':
            # Add the checkout URL and local
            r.append(spec.url)
            # Add the name of the local folder to checkout to, if it is
            # different from the last part of the URL.
            if spec.name: r.append(spec.name)
        elif command == 'update':
            # Add the local folder that must be updated
            r.append(str(local))
        # Add parameters for running commands in a non interactive mode
        r.append(Subversion.nonInteractive)
        # Add authentication-related parameters if found
        if spec.login:
            r.append('--username %s --password %s' % (spec.login,spec.password))
        return ' '.join(r), local

    def checkout(self, path):
        '''Checkout a copy of code retrieved via p_self.specifier, in some local
           p_path.'''
        # Remember the current working directory: we will chdir into p_path
        curdir = os.getcwd()
        os.chdir(path)
        # Build the svn command
        command, local = self.build('checkout', path)
        # Execute the command
        print('Executing %s...' % command)
        os.system(command)
        # chdir to the initial working directory
        os.chdir(curdir)
        # Return the path to the checkouted local copy
        return local
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import importlib.util

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def importModule(name, fileName):
    '''Imports module p_name given its absolute file p_name'''
    spec = importlib.util.spec_from_file_location(name, fileName)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)
    return mod

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Object:
    '''At every place we need an object, but without any requirement on its
       class (methods, attributes,...) we will use this minimalist class.'''

    def __init__(self, **fields):
        for k, v in fields.items(): setattr(self, k, v)

    def __repr__(self):
        '''A compact, string representation of this object for debugging
           purposes.'''
        r = '<O '
        for name, value in self.__dict__.items():
            # Avoid infinite recursion if p_self it auto-referenced
            if value == self: continue
            v = value
            if hasattr(v, '__repr__'):
                try:
                    v = v.__repr__()
                except TypeError:
                    pass
            try:
                r += '%s=%s ' % (name, v)
            except UnicodeDecodeError:
                r += '%s=<encoding problem> ' % name
        return r.strip() + '>'

    def __bool__(self): return bool(self.__dict__)
    def d(self): return self.__dict__
    def get(self, name, default=None): return getattr(self, name, default)
    def __contains__(self, k): return k in self.__dict__
    def keys(self): return self.__dict__.keys()
    def values(self): return self.__dict__.values()
    def items(self): return self.__dict__.items()

    def __setitem__(self, k, v):
        '''Dict-like attribute set'''
        self.__dict__[k] = v

    def __getitem__(self, k):
        '''Dict-like access self[k] must return None if key p_k doesn't exist'''
        return getattr(self, k, None)

    def __delitem__(self, k):
        '''Dict-like attribute removal'''
        del(self.__dict__[k])

    def __getattr__(self, name):
        '''Object access o.<name> must return None if attribute p_name does not
           exist.'''
        return

    def __eq__(self, other):
        '''Equality between objects is, like standard Python dicts, based on
           equality of all their attributes and values.'''
        if isinstance(other, Object):
            return self.__dict__ == other.__dict__
        return False

    def update(self, other):
        '''Set information from p_other (another Object instance or a dict) into
           p_self.'''
        other = other.__dict__ if isinstance(other, Object) else other
        for k, v in other.items(): setattr(self, k, v)

    def clone(self):
        '''Creates a clone from p_self'''
        r = Object()
        r.update(self)
        return r

    # Allow this highly manipulated Object class to be picklable
    def __getstate__(self): return self.__dict__
    def __setstate__(self, state): self.__dict__.update(state)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Hack:
    '''This class proposes methods for patching some existing code with
       alternative methods.'''

    @staticmethod
    def patch(method, replacement, class_=None):
        '''This method replaces m_method with a p_replacement method, but
           keeps p_method on its class under name
           "_base_<initial_method_name>_". In the patched method, one may use
           Hack.base to call the base method. If p_method is static, you must
           specify its class in p_class_.'''
        # Get the class on which the surgery will take place
        isStatic = class_
        class_ = class_ or method.im_class
        # On this class, store m_method under its "base" name
        name = isStatic and method.func_name or method.im_func.__name__
        baseName = '_base_%s_' % name
        if isStatic:
            # If "staticmethod" isn't called hereafter, the static functions
            # will be wrapped in methods.
            method = staticmethod(method)
            replacement = staticmethod(replacement)
        setattr(class_, baseName, method)
        setattr(class_, name, replacement)

    @staticmethod
    def base(method, class_=None):
        '''Allows to call the base (replaced) method. If p_method is static,
           you must specify its p_class_.'''
        isStatic = class_
        class_ = class_ or method.im_class
        name = isStatic and method.func_name or method.im_func.__name__
        return getattr(class_, '_base_%s_' % name)

    @staticmethod
    def inject(patchClass, class_, verbose=False):
        '''Injects any method or attribute from p_patchClass into class_.'''
        # As a preamble, inject methods and attributes from p_patchClass's base
        # classes, if any.
        for base in patchClass.__bases__:
            Hack.inject(base, class_, verbose=verbose)
        patched = []
        added = []
        # Inject p_patchClass' own methods and attributes
        for name, attr in patchClass.__dict__.items():
            if name.startswith('__'): continue # Ignore special methods
            # Unwrap functions from static methods
            className = attr.__class__.__name__
            if className == 'staticmethod':
                attr = attr.__get__(attr)
                static = True
            else:
                static = False
            # Is this name already defined on p_class_ ?
            if hasattr(class_, name):
                hasAttr = True
                classAttr = getattr(class_, name)
            else:
                hasAttr = False
                classAttr = None
            if hasAttr and (className != 'type') and \
               callable(attr) and callable(classAttr):
                # Patch this method via Hack.patch
                if static:
                    Hack.patch(classAttr, attr, class_)
                else:
                    Hack.patch(classAttr, attr)
                patched.append(name)
            else:
                # Simply replace the static attr or add the new static
                # attribute or method.
                setattr(class_, name, attr)
                added.append(name)
        if verbose:
            pName = patchClass.__name__
            cName = class_.__name__
            print('%d method(s) patched from %s to %s (%s)' % \
                  (len(patched), pName, cName, str(patched)))
            print('%d method(s) and/or attribute(s) added from %s to %s (%s)'%\
                  (len(added), pName, cName, str(added)))

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def notFromPortlet(tool):
    '''When set in a root class' attribute "createVia", this function prevents
       instances being created from the portlet.'''

    # This is practical if you want to get the facilities provided by the
    # "portlet zone" for a class (ie, all search facilities) but without
    # allowing people people to create such classes at the root level.
    try:
        return 'form' if not tool.traversal.context.rootClasses else None
    except AttributeError:
        pass
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
from appy.px import Px
from appy.fields import Field

# ------------------------------------------------------------------------------
class Color(Field):
    '''Field representing a color'''

    pxView = pxCell = Px('''
     <input type="color" disabled=":True" value=":value or '#000000'"
            name=":name" id=":name"/>''')

    pxEdit = pxSearch = Px('''
     <input type="color" id=":name" name=":name" size=":field.width"
            value=":field.getInputValue(inRequest, requestValue, value)"/>
     <script if="hostLayout">:'prepareForAjaxSave(%s,%s,%s,%s)' % \
      (q(name),q(obj.id),q(obj.url),q(hostLayout))</script>''')

    def __init__(self, validator=None, multiplicity=(0,1), default=None,
      defaultOnEdit=None, show=True, page='main', group=None, layouts=None,
      move=0, indexed=False, mustIndex=True, indexValue=None, searchable=False,
      specificReadPermission=False, specificWritePermission=False, width=5,
      height=None, maxChars=13, colspan=1, master=None, masterValue=None,
      focus=False, historized=False, mapping=None, generateLabel=None,
      label=None, sdefault=('',''), scolspan=1, swidth=None, sheight=None,
      persist=True, inlineEdit=False, view=None, cell=None, edit=None, xml=None,
      translations=None):
        # Call the base constructor
        Field.__init__(self, validator, multiplicity, default, defaultOnEdit,
          show, page, group, layouts, move, indexed, mustIndex, indexValue,
          searchable, specificReadPermission, specificWritePermission, width,
          height, maxChars, colspan, master, masterValue, focus, historized,
          mapping, generateLabel, label, sdefault, scolspan, swidth, sheight,
          persist, inlineEdit, view, cell, edit, xml, translations)
        # Define the corresponding Python type for values stored for this type
        self.pythonType = str
# ------------------------------------------------------------------------------



'''Appy is the simpliest way to build complex webapps'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import pathlib
# Store here the path to the Appy root package, it is often requested
path = pathlib.Path(__file__).parent

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Config:
    '''Root of all configuration options for your app'''

    # These options are those managed by the app developer: they are not meant
    # to be edited during the app lifetime by end users. For such "end-user"
    # configuration options, you must extend the appy.model.tool.Tool class,
    # designed for that purpose. In short, this Config file represents the "RAM"
    # configuration, while the unique appy.model.tool.Tool instance within every
    # app contains its "DB" configuration.

    # In your app/__init__.py, create a class named "Config" that inherits from
    # this one and will override some of the atttibutes defined here, ie:

    # import appy
    # class Config(appy.Config):
    #     someAttribute = "someValue"

    # If "someAttribute" is not a standard Appy attribute, this is a way to add
    # your own configuration attributes.

    # If you want to modify existing attributes, like model configuration or
    # user interface configuration (that, if you have used appy/bin/make to
    # generate your app, are already instantiated in attributes "model" and
    # "ui"), after the attribute definition, modify it like this:

    # class Config(appy.Config):
    #     ...
    #     ui.languages = ('en', 'fr')
    #     model.rootClasses = ['MyClass']

    # Place here a appy.server.Config instance defining the configuration
    # of the Appy HTTP server.
    server = None
    # Place here a appy.server.guard.Config instance defining security options
    security = None
    # Place here a appy.database.Config instance defining database options
    database = None
    # Place here a appy.database.log.Config instance defining logging options
    log = None
    # Place here a appy.model.Config instance defining the application model
    model = None
    # Place here a appy.ui.Config instance defining user-interface options
    ui = None
    # When using a SMTP mail server for sending emails from your app, place an
    # instance of class appy.utils.mail.Config in the field below.
    mail = None
    # Place here a appy.server.scheduler.Config instance, defining cron-like
    # actions being automatically triggered.
    jobs = None
    # Place here an instance of appy.server.backup.Config if you want to backup
    # data and logs from a Appy site.
    backup = None
    # Place here a appy.deploy.Config instance defining how to deploy this app
    # on distant servers.
    deploy = None
    # When using Google analytics, specify here the Analytics ID
    googleAnalyticsId = None
    # When using Ogone, place an instance of appy.model.fields.ogone.Config in
    # the field below.
    ogone = None
    # When using POD fields for producing documents with appy.pod, place here an
    # instance of appy.model.fields.pod.Config
    pod = None
    # When the app has an extension, the name of this latter will be stored
    # in the following attribute. Do not set it directly: it must be done via
    # method m_declareExt below. If you use script appy/bin/make to create an
    # ext, a call to declareExt will be present in the generated ext's
    # __init__.py file.
    ext = None
    # An Appy site may communicate with peer sites. Defining peer sites is done
    # by placing, in the following attribute, an instance of appy.peer.Config.
    peers = None

    @classmethod
    def declareExt(class_, path):
        '''Declares an extension to this app, whose path is p_path'''
        name = path.name
        class_.ext = name
        # Add the ext's "static" folder to the static map
        class_.server.static.map[name] = path / 'static'

    @classmethod
    def check(self):
        '''Ensures the config is valid. Called at server startup'''
        self.server.check()
        self.server.static.check()
        self.security.check()
        if self.jobs:
            self.jobs.check()
        if self.backup:
            self.backup.check()
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''This script allows to create a brand new ready-to-use Plone/Zone instance.
   As prerequisite, you must have installed Plone through the Unifier installer
   available at http://plone.org.'''

# ------------------------------------------------------------------------------
import os, os.path, sys, shutil, re
from optparse import OptionParser
from appy.shared.utils import cleanFolder, copyFolder, executeCommand
from appy.shared.packaging import loStart, zopeConf

# ------------------------------------------------------------------------------
class NewError(Exception): pass
ERROR_CODE = 1
WRONG_NB_OF_ARGS = 'Wrong number of args.'
WRONG_PLONE_VERSION = 'Plone version must be among %s.'
WRONG_PLONE_PATH = 'Path "%s" is not an existing folder.'
PYTHON_NOT_FOUND = 'Python interpreter was not found in "%s". Are you sure ' \
    'we are in the folder hierarchy created by the Plone installer?'
PYTHON_EXE_NOT_FOUND = '"%s" does not exist.'
MKZOPE_NOT_FOUND = 'Script mkzopeinstance.py not found in "%s and ' \
    'subfolders. Are you sure we are in the folder hierarchy created by ' \
    'the Plone installer?'
WRONG_INSTANCE_PATH = '"%s" must be an existing folder for creating the ' \
    'instance in it.'

# zopectl template file for a pure Zope instance -------------------------------
zopeCtl = '''#!/bin/sh
PYTHON="/usr/lib/zope2.12/bin/python"
INSTANCE_HOME="%s"
CONFIG_FILE="$INSTANCE_HOME/etc/zope.conf"
ZDCTL="/usr/lib/zope2.12/bin/zopectl"
export INSTANCE_HOME
export PYTHON
exec "$ZDCTL" -C "$CONFIG_FILE" "$@"
'''

# runzope template file for a pure Zope instance -------------------------------
runZope = '''#!/bin/sh
INSTANCE_HOME="%s"
CONFIG_FILE="$INSTANCE_HOME/etc/zope.conf"
ZOPE_RUN="/usr/lib/zope2.12/bin/runzope"
export INSTANCE_HOME
exec "$ZOPE_RUN" -C "$CONFIG_FILE" "$@"
'''

# zopectl template for a Plone (4) Zope instance -------------------------------
zopeCtlPlone = '''#!/bin/sh
PYTHON="%s"
INSTANCE_HOME="%s"
CONFIG_FILE="$INSTANCE_HOME/etc/zope.conf"
PYTHONPATH="$INSTANCE_HOME/lib/python"
ZDCTL="%s/Zope2/Startup/zopectl.py"
export INSTANCE_HOME
export PYTHON
export PYTHONPATH
exec "$PYTHON" "$ZDCTL" -C "$CONFIG_FILE" "$@"
'''

# runzope template for a Plone (4) Zope instance -------------------------------
runZopePlone = '''#! /bin/sh
PYTHON="%s"
INSTANCE_HOME="%s"
CONFIG_FILE="$INSTANCE_HOME/etc/zope.conf"
PYTHONPATH="$INSTANCE_HOME/lib/python"
ZOPE_RUN="%s/Zope2/Startup/run.py"
export INSTANCE_HOME
export PYTHON
export PYTHONPATH
exec "$PYTHON" "$ZOPE_RUN" -C "$CONFIG_FILE" "$@"
'''

# Patch to apply to file pkg_resources.py in a Plone4 Zope instance ------------
pkgResourcesPatch = '''import os, os.path
productsFolder = os.path.join(os.environ["INSTANCE_HOME"], "Products")
for name in os.listdir(productsFolder):
    if os.path.isdir(os.path.join(productsFolder, name)):
        if name not in appyVersions:
            appyVersions[name] = "1.0"
            appyVersions['Products.%s' % name] = "1.0"

def getAppyVersion(req, location):
    global appyVersions
    if req.project_name not in appyVersions:
        raise DistributionNotFound(req)
    return Distribution(project_name=req.project_name,
                        version=appyVersions[req.project_name],
                        platform='linux2', location=location)
'''

# ------------------------------------------------------------------------------
class ZopeInstanceCreator:
    '''This class allows to create a Zope instance. It makes the assumption that
       Zope was installed via the Debian package zope2.12.'''

    def __init__(self, instancePath):
        self.instancePath = instancePath

    def run(self):
        # Create the instance folder hierarchy
        if not os.path.exists(self.instancePath):
            os.makedirs(self.instancePath)
        curdir = os.getcwd()
        # Create bin/zopectl
        os.chdir(self.instancePath)
        os.mkdir('bin')
        f = file('bin/zopectl', 'w')
        f.write(zopeCtl % self.instancePath)
        f.close()
        os.chmod('bin/zopectl', 0744) # Make it executable by owner.
        # Create bin/runzope
        f = file('bin/runzope', 'w')
        f.write(runZope % self.instancePath)
        f.close()
        os.chmod('bin/runzope', 0744) # Make it executable by owner.
        # Create bin/startlo
        f = file('bin/startlo', 'w')
        f.write(loStart)
        f.close()
        os.chmod('bin/startlo', 0744) # Make it executable by owner.
        # Create etc/zope.conf
        os.mkdir('etc')
        f = file('etc/zope.conf', 'w')
        f.write(zopeConf % (self.instancePath, '%s/var' % self.instancePath,
                            '%s/log' % self.instancePath, '8080', ''))
        f.close()
        # Create other folders
        for name in ('Extensions', 'log', 'Products', 'var'): os.mkdir(name)
        f = file('Products/__init__.py', 'w')
        f.write('#Makes me a Python package.\n')
        f.close()
        # Create 'inituser' file with admin password
        import binascii
        try:
            from hashlib import sha1 as sha
        except:
            from sha import new as sha
        f = open('inituser', 'w')
        password = binascii.b2a_base64(sha('admin').digest())[:-1]
        f.write('admin:{SHA}%s\n' % password)
        f.close()
        os.chmod('inituser', 0644)
        # User "zope" must own this instance
        executeCommand('chown', '-R', 'zope', self.instancePath)
        print('Zope instance created in %s.' % self.instancePath)
        os.chdir(curdir)

# ------------------------------------------------------------------------------
class NewScript:
    '''usage: %prog ploneVersion plonePath instancePath

       "ploneVersion"  can be plone25, plone30, plone3x, plone4 or zope
                       (plone3x represents Plone 3.2.x, Plone 3.3.5...)
       
       "plonePath"     is the (absolute) path to your plone (or zope)
                       installation. Plone 2.5 and 3.0 are typically installed
                       in /opt/Plone-x.x.x, while Plone 3 > 3.0 is typically
                       installed in in /usr/local/Plone.
       "instancePath"  is the (absolute) path where you want to create your
                       instance (should not already exist).'''
    ploneVersions = ('plone25', 'plone30', 'plone3x', 'plone4', 'zope')

    def installPlone25or30Stuff(self, linksForProducts):
        '''Here, we will copy all Plone2-related stuff in the Zope instance
           we've created, to get a full Plone-ready Zope instance. If
           p_linksForProducts is True, we do not perform a real copy: we will
           create symlinks to products lying within Plone installer files.'''
        j = os.path.join
        if self.ploneVersion == 'plone25':
            sourceFolders = ('zeocluster/Products',)
        else:
            sourceFolders = ('zinstance/Products', 'zinstance/lib/python')
        for sourceFolder in sourceFolders:
            sourceBase = j(self.plonePath, sourceFolder)
            destBase = j(self.instancePath,
                         sourceFolder[sourceFolder.find('/')+1:])
            for name in os.listdir(sourceBase):
                folderName = j(sourceBase, name)
                if os.path.isdir(folderName):
                    destFolder = j(destBase, name)
                    # This is a Plone product. Copy it to the instance.
                    if linksForProducts:
                        # Create a symlink to this product in the instance
                        executeCommand(['ln', '-s', folderName, destFolder])
                    else:
                        # Copy thre product into the instance
                        copyFolder(folderName, destFolder)

    filesToPatch = ('meta.zcml', 'configure.zcml', 'overrides.zcml')
    patchRex = re.compile('<includePlugins.*?/>', re.S)
    def patchPlone3x(self):
        '''Auto-proclaimed ugly code in z3c forces us to patch some files
           in Products.CMFPlone because these guys make the assumption that
           "plone.*" packages are within eggs when they've implemented their
           ZCML directives "includePlugins" and "includePluginsOverrides".
           So in this method, I remove every call to those directives in
           CMFPlone files. It does not seem to affect Plone behaviour. Indeed,
           these directives seem to be useful only when adding sad (ie, non
           Appy) Plone plug-ins.'''
        j = os.path.join
        ploneFolder = os.path.join(self.productsFolder, 'CMFPlone')
        # Patch files
        for fileName in self.filesToPatch:
            filePath = os.path.join(ploneFolder, fileName)
            f = file(filePath)
            fileContent = f.read()
            f.close()
            f = file(filePath, 'w')
            f.write(self.patchRex.sub('<!--Del. includePlugins-->',fileContent))
            f.close()

    missingIncludes = ('plone.app.upgrade', 'plonetheme.sunburst',
                       'plonetheme.classic')
    def patchPlone4(self, versions):
        '''Patches Plone 4 that can't live without buildout as-is.'''
        self.patchPlone3x() # We still need this for Plone 4 as well.
        # bin/zopectl
        content = zopeCtlPlone % (self.pythonPath, self.instancePath,
                                   self.zopePath)
        f = file('%s/bin/zopectl' % self.instancePath, 'w')
        f.write(content)
        f.close()
        # bin/runzope
        content = runZopePlone % (self.pythonPath, self.instancePath,
                                  self.zopePath)
        f = file('%s/bin/runzope' % self.instancePath, 'w')
        f.write(content)
        f.close()
        j = os.path.join
        # As eggs have been deleted, versions of components are lost. Reify
        # them from p_versions.
        dVersions = ['"%s":"%s"' % (n, v) for n, v in versions.iteritems()]
        sVersions = 'appyVersions = {' + ','.join(dVersions) + '}'
        codeFile = "%s/pkg_resources.py" % self.libFolder
        f = file(codeFile)
        content = f.read().replace("raise DistributionNotFound(req)",
                         "dist = getAppyVersion(req, '%s')" % self.instancePath)
        content = sVersions + '\n' + pkgResourcesPatch + '\n' + content
        f.close()
        f = file(codeFile, 'w')
        f.write(content)
        f.close()
        # Some 'include' directives must be added with our install.
        configPlone = j(self.productsFolder, 'CMFPlone', 'configure.zcml')
        f = file(configPlone)
        content = f.read()
        f.close()
        missing = ''
        for missingInclude in self.missingIncludes:
            missing += '  <include package="%s"/>\n' % missingInclude
        content = content.replace('</configure>', '%s\n</configure>' % missing)
        f = file(configPlone, 'w')
        f.write(content)
        f.close()

    def copyEggs(self):
        '''Copy content of eggs into the Zope instance. This method also
           retrieves every egg version and returns a dict {s_egg:s_version}.'''
        j = os.path.join
        eggsFolder = j(self.plonePath, 'buildout-cache/eggs')
        res = {}
        for name in os.listdir(eggsFolder):
            if name == 'EGG-INFO': continue
            splittedName = name.split('-')
            res[splittedName[0]] = splittedName[1]
            if splittedName[0].startswith('Products.'):
                res[splittedName[0][9:]] = splittedName[1]
            absName = j(eggsFolder, name)
            # Copy every file or sub-folder into self.libFolder or
            # self.productsFolder.
            for fileName in os.listdir(absName):
                absFileName = j(absName, fileName)
                if fileName == 'Products' and not name.startswith('Zope2-'):
                    # Copy every sub-folder into self.productsFolder
                    for folderName in os.listdir(absFileName):
                        absFolder = j(absFileName, folderName)
                        if not os.path.isdir(absFolder): continue
                        copyFolder(absFolder, j(self.productsFolder,folderName))
                elif os.path.isdir(absFileName):
                    copyFolder(absFileName, j(self.libFolder, fileName))
                else:
                    shutil.copy(absFileName, self.libFolder)
        return res

    def createInstance(self, linksForProducts):
        '''Calls the Zope script that allows to create a Zope instance and copy
           into it all the Plone packages and products.'''
        j = os.path.join
        # Find the Python interpreter running Zope
        for elem in os.listdir(self.plonePath):
            pythonPath = None
            elemPath = j(self.plonePath, elem)
            if elem.startswith('Python-') and os.path.isdir(elemPath):
                pythonPath = elemPath + '/bin/python'
                if not os.path.exists(pythonPath):
                    raise NewError(PYTHON_EXE_NOT_FOUND % pythonPath)
                break
        if not pythonPath:
            raise NewError(PYTHON_NOT_FOUND % self.plonePath)
        self.pythonPath = pythonPath
        # Find the Zope script mkzopeinstance.py and Zope itself
        makeInstancePath = None
        self.zopePath = None
        for dirname, dirs, files in os.walk(self.plonePath):
            # Find Zope
            for folderName in dirs:
                if folderName.startswith('Zope2-'):
                    self.zopePath = j(dirname, folderName)
            # Find mkzopeinstance
            for fileName in files:
                if fileName == 'mkzopeinstance.py':
                    if self.ploneVersion == 'plone4':
                        makeInstancePath = j(dirname, fileName)
                    else:
                        if ('/buildout-cache/' not in dirname):
                            makeInstancePath = j(dirname, fileName)
        if not makeInstancePath:
            raise NewError(MKZOPE_NOT_FOUND % self.plonePath)
        # Execute mkzopeinstance.py with the right Python interpreter.
        # For Plone4, we will call it later.
        cmd = [pythonPath, makeInstancePath, '-d', self.instancePath]
        if self.ploneVersion != 'plone4':
            print(cmd)
            executeCommand(cmd)
        # Now, make the instance Plone-ready
        action = 'Copying'
        if linksForProducts:
            action = 'Symlinking'
        print('%s Plone stuff in the Zope instance...' % action)
        if self.ploneVersion in ('plone25', 'plone30'):
            self.installPlone25or30Stuff(linksForProducts)
        elif self.ploneVersion in ('plone3x', 'plone4'):
            versions = self.copyEggs()
            if self.ploneVersion == 'plone3x':
                self.patchPlone3x()
            elif self.ploneVersion == 'plone4':
                # Create the Zope instance
                os.environ['PYTHONPATH'] = '%s:%s' % \
                    (j(self.instancePath,'Products'),
                     j(self.instancePath, 'lib/python'))
                print(cmd)
                executeCommand(cmd)
                self.patchPlone4(versions)
        # Remove .bat files under Linux
        if os.name == 'posix':
            cleanFolder(j(self.instancePath, 'bin'), exts=('.bat',))

    def manageArgs(self, args):
        '''Ensures that the script was called with the right parameters.'''
        if len(args) != 3: raise NewError(WRONG_NB_OF_ARGS)
        self.ploneVersion, self.plonePath, self.instancePath = args
        # Add some more folder definitions
        j = os.path.join
        self.productsFolder = j(self.instancePath, 'Products')
        self.libFolder = j(self.instancePath, 'lib/python')
        # Check Plone version
        if self.ploneVersion not in self.ploneVersions:
            raise NewError(WRONG_PLONE_VERSION % str(self.ploneVersions))
        # Check Plone path
        if not os.path.exists(self.plonePath) \
           or not os.path.isdir(self.plonePath):
            raise NewError(WRONG_PLONE_PATH % self.plonePath)
        # Check instance path
        parentFolder = os.path.dirname(self.instancePath)
        if not os.path.exists(parentFolder) or not os.path.isdir(parentFolder):
            raise NewError(WRONG_INSTANCE_PATH % parentFolder)

    def run(self):
        optParser = OptionParser(usage=NewScript.__doc__)
        optParser.add_option("-l", "--links", action="store_true",
            help="[Linux, plone25 or plone30 only] Within the created " \
                 "instance, symlinks to Products lying within the Plone " \
                 "installer files are created instead of copying them into " \
                 "the instance. This avoids duplicating the Products source " \
                 "code and is interesting if you create a lot of Zope " \
                 "instances.")
        (options, args) = optParser.parse_args()
        linksForProducts = options.links
        try:
            self.manageArgs(args)
            if self.ploneVersion != 'zope':
                print('Creating new %s instance...' % self.ploneVersion)
                self.createInstance(linksForProducts)
            else:
                ZopeInstanceCreator(self.instancePath).run()
        except NewError, ne:
            optParser.print_help()
            sys.stderr.write(str(ne))
            sys.stderr.write('\n')
            sys.exit(ERROR_CODE)

# ------------------------------------------------------------------------------
if __name__ == '__main__':
    NewScript().run()
# ------------------------------------------------------------------------------



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.px import Px
from appy.ui import css
from appy.model.base import Base
from appy.ui.layout import Layouts
from appy.model.fields.ref import Ref
from appy.model.fields.file import File
from appy.model.fields.dict import Dict
from appy.model.utils import Object as O
from appy.model.document import Document
from appy.model.fields.phase import Page
from appy.model.fields.color import Color
from appy.model.fields.group import Group
from appy.model.fields.string import String
from appy.model.fields.boolean import Boolean
from appy.model.fields.integer import Integer
from appy.model.fields.computed import Computed
from appy.model.workflow.standard import Anonymous
from appy.model.fields.select import Select, Selection

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Carousel(Base):
    '''A carousel is a UI element allowing to cycle through documents'''

    # By default, carousels are public, and not indexed
    workflow = Anonymous
    indexable = False
    listColumns = ('title', 'maxWidth', 'autoSwitch', 'switchInterval')
    pageListColumns = ('title', 'state')

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #                            Main parameters
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # Fields in the main page are rendered in a grid group
    mainGroup = Group('main', ['400em', ''], style='grid', hasLabel=False)

    pc = {'label': 'Carousel', 'group': mainGroup}

    # Maximum width for documents in the carousel (in pixels)
    maxWidth = Integer(default=2277, multiplicity=(1,1), layouts=Layouts.gd,
                       alignOnEdit='center', **pc)

    # Maximum height for the carousel (bottom image excluded, see below). Set it
    # to 'none' if you don't want to define this maximum height. If 'none', the
    # real height will depend on the inner images. Whatever you choose, images'
    # width/height ratios are always kept.
    maxHeight = String(default='620px', multiplicity=(1,1), width=5,
                       layouts=Layouts.gd, **pc)

    # Must automatic switch be enabled ?
    autoSwitch = Boolean(default=True, **pc)
    # Duration between 2 switches (in seconds)
    switchInterval = Integer(default=5, multiplicity=(1,1), width=2,
                             alignOnEdit='center', **pc)

    # Default attributes for image's elements
    def listFonts(self):
        '''Lists the available Google fonts as specified in the config'''
        return [(font, font) for font in self.config.ui.googleFonts]

    def showFonts(self):
        '''Do not show font-related field if there is not Google font to
           choose.'''
        return bool(self.config.ui.googleFonts)

    defaultElemAttributes = {
      'title'     : O(color='#ffffff', font='', fontSize='3.5vw',
                      fontWeight='normal'),
      'underTitle': O(color='#afb0b9', font='', fontSize='1.5vw',
                      fontWeight='normal'),
      'buttonA'   : O(color='#ffffff', font='', fontSize='1vw',
                      fontWeight='bold'),
      'buttonB'   : O(color='#ffffff', font='', fontSize='1vw',
                      fontWeight='bold')
    }

    sub = (
      ('color', Color()), # Text color
      ('font', Select(validator=Selection(listFonts), show=showFonts)),
      ('fontSize', String(width=4, multiplicity=(1,1))),
      ('fontWeight', Select(validator=Selection(
        lambda o: [(w, o.translate('fw_%s' % w)) for w in css.fontWeigths]),
        default='normal', multiplicity=(1,1))),
      ('fontStyle', Select(validator=Selection(
        lambda o: [(s, o.translate('fs_%s' % s)) for s in css.fontStyles]),
        default='normal', multiplicity=(1,1)))
    )

    elemAttributes = Dict(lambda o: Document.listElementTypes(o), sub,
                          default=defaultElemAttributes, **pc)

    # Default attributes for links that can be set on image elements
    subl = (
      ('params', String()), # Default parameters to add to every link URL
      ('target', String(width=7)), # Link's target (_blank, iframe,...)
      ('onClick', String()), # Optional value for "onClick" attribute
    )

    linkAttributes = Dict(lambda o: Document.listElementTypes(o), subl, **pc)

    # Default background colors for buttons
    buttonBgColorA = Color(default='#88c9b5', **pc)
    buttonBgColorB = Color(default='#fcc100', **pc)
    # Map buttons types to their backgrounds
    backgroundFields = {'buttonA': 'buttonBgColorA',
                        'buttonB': 'buttonBgColorB'}

    # Colors for the controls allowing to (un)select a given image
    selectedColor = Color(default='#fcc100', **pc)
    unselectedColor = Color(default='#ffffff', **pc)

    # Image to use as bottom border
    bottomImage = File(isImage=True, resize=True, width=lambda o: o.maxWidth,
                       viewWidth='700px', layouts=Layouts.gd, cache=True, **pc)

    def getCssName(self, elemType):
        '''Gets the name of the CSS class corresponding to p_elemType'''
        return 'cl_%d_%s' % (self.iid, elemType)

    def getElementStyles(self):
        '''Generates the CSS classes to apply to elements shown on top of
           carousel's images.'''
        r = []
        attributes = self.elemAttributes
        for type in Document.elementTypes:
            # Compute background color when appropriate
            bg = Carousel.backgroundFields.get(type)
            bg = (';background-color:%s' % getattr(self, bg)) if bg else ''
            padding = ';padding:0.7em 1.7em' if bg else ''
            # Compute other CSS attributes
            attrs = attributes[type]
            font = ('"%s"' % attrs.font) if attrs.font else 'sans-serif'
            className = self.getCssName(type)
            css = '.%s {color:%s;font-family:%s;font-size:%s;font-weight:%s;' \
                  'font-style:%s%s%s}' % \
                  (className, attrs.color, font, attrs.fontSize,
                   attrs.fontWeight, attrs.fontStyle, bg, padding)
            # Add another rule for coloring the text of "a" sub-tags
            cssA = '.%s a, .%s a:visited {color:%s}' % \
                   (className, className, attrs.color)
            r.append(css)
            r.append(cssA)
        return '\n'.join(r)

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #                           Inner images
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # The documents being part of this carousel
    del(pc['group'])
    documents = Ref(Document, add=True, link=False, multiplicity=(0,None),
      composite=True, back=Ref(attribute='carousel', show=False),
      showHeaders=True, shownInfo=Document.listColumns, actionsDisplay='inline',
      page=Page('documents',
                show=lambda o:'view' if o.allows('write') else None),
      viaPopup=False, rowAlign='middle', **pc)

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #                          PX rendering
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    pxView = Px('''
     <!-- Inject styles dynamically generated for every element -->
     <style>::o.getElementStyles()</style>
     <x if="o.isEmpty('documents')">-</x>
     <div if="not o.isEmpty('documents')"
          var2="carouselId='cl_%d_%s' % (o.iid, field.name);
                count=o.countRefs('documents')"
          class="carousel" id=":carouselId" data-count=":count" data-current="1"
          data-selected-color=":o.selectedColor" 
          data-unselected-color=":o.unselectedColor"
          data-auto-switch=":str(o.autoSwitch)"
          data-switch-interval=":o.switchInterval" data-timer-id="">

      <!-- The images -->
      <div for="image in o.documents"
           var2="i=loop.image.nb"
           id=":'%s_%d' % (carouselId, i+1)"
           style=":'opacity:%s' % ('1' if i == 0 else '0')">
       <img src=":'%s/file/download' % image.url"
            style=":'max-height:%s' % o.maxHeight"/>

       <!-- The elements to superpose on the image -->
       <div class="canvas">
         <x for="element in image.getElements(o)">::element</x>
       </div>
      </div>

      <!-- The controls for selecting a specific element -->
      <svg width=":20+((count-1)*30)" height="20">
       <circle for="image in o.documents" var2="i=loop.image.nb"
               id=":'%s_circle_%d' % (carouselId, i+1)"
               class="clickable" cx=":10 + (i*30)" cy="10" r="10"
               onclick=":'carouselSelect(%s,%s,%d,true)' % \
                          (q(carouselId), q(i+1), count)"
               fill=":o.selectedColor if i == 0 else o.unselectedColor" />
      </svg>
      <script>:'carouselStart(%s)' % q(carouselId)</script>
     </div>
     <!-- Render the bottom image when present -->
     <img if="not o.isEmpty('bottomImage')" class="carouselFooter"
          src=":'%s/bottomImage/download' % o.url"/>''',

     css='''
      .carousel { position: relative }
      .carousel > div { transition: opacity 2s }
      .carousel > div:nth-child(n+2) { position: absolute; top: 0; left: 0;
                                       width: 100% }
      .carousel img { width: 100%; height: auto; object-fit: cover }
      .carousel svg { position: absolute; top: 90%; left: 50%; z-index: 2;
                      transform: translate(-50%,0%) }
      .carousel .canvas { top: 0; left: 0; position: absolute;
                          width: 90%; height: 100%; z-index: 2 }
      .carousel .canvas > div { position: absolute }
      .carouselFooter { width: 100%; height: auto; vertical-align: top }
     ''',

     js='''
      function getCarouselPart(carouselId, i, isImage) {
        /* Get the carousel part numbered p_i: the ith image if p_isImage is
           True, the selection circle else. */
        var sep = (isImage)? '_': '_circle_';
        return document.getElementById(carouselId + sep + i.toString());
      }
      function carouselSelect(carouselId, nb, count, stopTimer) {
        // Make element numbered p_nb visible in the carousel
        var nb=parseInt(nb),
            selected = getCarouselPart(carouselId, nb, true),
            carousel=document.getElementById(carouselId);
        // Stop the timer when appropriate
        if (stopTimer) {
          var timerId = carousel.dataset.timerId;
          if (timerId) clearInterval(parseInt(timerId));
        }
        // Do nothing more if the selected image is already visible
        if (selected.style.opacity == 1) return;
        // Find the element being currently shown
        var part=null, circle=null;
        for (var i=0; i<count; i++) {
          part = getCarouselPart(carouselId, i+1, true);
          if (part.style.opacity == 1) {
            // Switch the color of their corresponding circle
            circle = getCarouselPart(carouselId, i+1, false);
            circle.setAttribute('fill', carousel.dataset.unselectedColor);
            circle = getCarouselPart(carouselId, nb, false);
            circle.setAttribute('fill', carousel.dataset.selectedColor);
            // Switch visibility of parts
            selected.style.opacity = 1;
            part.style.opacity = 0;
            // Remember the new current image
            carousel.dataset.current = nb.toString();
            return;
          }
        }
      }
      function carouselSwitch(carouselId) {
        // Switch from the currently shown image to the next one
        var carousel = document.getElementById(carouselId),
            count = parseInt(carousel.dataset.count),
            current = parseInt(carousel.dataset.current),
            next=(current==count)? 1: current+1;
        carouselSelect(carouselId, next, count);
      }
      function carouselStart(carouselId) {
        var carousel = document.getElementById(carouselId),
            auto = carousel.dataset.autoSwitch == 'True',
            count = parseInt(carousel.dataset.count);
        // Do not launch it if automatic switch is disabled
        if (!auto || (count == 0)) return;
        var interval = parseInt(carousel.dataset.switchInterval),
            timerId = setInterval(
              function() {carouselSwitch(carouselId);}, interval*1000);
        carousel.dataset.timerId = timerId.toString();
      }''')

    # A field allowing to preview the carousel in one of its pages
    preview = Computed(method=pxView, page=Page('preview', show='view'),
                       layouts=Layouts.w, **pc)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''Import data from a Appy 0.x site to an Appy 1.x site'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.peer import Peer
from appy.peer.importer import Importer, UserImporter

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
NOT_IMPL_YET = 'Not implemented yet.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Importer0(Importer):
    '''Base object importer for Appy 0'''

    def setHistory(self):
        '''Imports Appy 0's history from p_self.distant and populate the
           corresponding Appy 1 data structure on p_self.local.'''
        history = self.local.history
        # History events in Appy 0 are chronologically sorted
        i = -1
        for event in self.distant.history:
            i += 1
            # Creator, created & modified will be set by m_setFields
            if i == 0: continue
            # Get parameters being common to all history events
            params = {'state': event.review_state, 'date': event.time,
                      'login': event.actor, 'comment': event.comments}
            action = event.action
            if action == '_datachange_':
                eventType = 'Change'
                # Convert the "changes" dict to Appy 1
                changes = {}
                for name, vals in event.changes.items():
                    changes[name] = vals[0]
                params['changes'] = changes
            elif action == '_datadelete_':
                print('Data-delete event', event)
                raise Exception(NOT_IMPL_YET)
            elif action == '_dataadd_':
                print('Data-add event', event)
                raise Exception(NOT_IMPL_YET)
            else:
                # A workflow transition
                params['transition'] = action
                eventType = 'Trigger'
            history.add(eventType, **params)

    def setLocalRoles(self):
        '''Sets local roles on p_self.local as defined on p_self.distant'''
        # Delete local role granted to technical user "system" when the local
        # object has been created.
        localRoles = self.local.localRoles
        localRoles.delete('system')
        for login, roles in self.distant.localRoles.items():
            localRoles.add(login, roles)

    def customize(self):
        '''Manage "creator", "created" and "modified" fields, being stored
           differently in Appy 0 and Appy 1.'''
        o = self.local
        d = self.distant
        o.setCreator(d.creator)
        o.history[-1].date = d.created
        o.history.modified = d.modified

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class UserImporter0(Importer0, UserImporter):
    '''Appy 0 User-specific importer'''

    def customize(self):
        '''Perform Appy 0 AND User-specific customization'''
        Importer0.customize(self)
        UserImporter.customize(self)

    @classmethod
    def getAction(class_, distant):
        '''Bypass Python's standard depth-first search inheritance and call
           UserImporter.getAction.'''
        return UserImporter.getAction(distant)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Appy0(Peer):
    '''Represents a peer Appy 0.x site'''

    # Declare the base object importer for Appy 0
    baseImporter = Importer0

    # In Appy 0, the ID of the tool is 'config' and not 'tool'
    distantToolName = 'config'

    def __init__(self, *args, **kwargs):
        # Call the base constructor
        super().__init__(*args, **kwargs)
        # Update importers
        self.importers['User'] = UserImporter0

    def getSearchUrl(self, tool, className):
        '''Return the URL allowing to retrieve URLs of all instances from a
           given p_className on a distant site.'''
        # Get the name of the distant class corresponding to this local
        # p_className.
        distantName = self.classNames.get(className) or className
        return '%s/config?do=searchAll&className=%s' % (self.url, distantName)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# -*- coding: utf-8 -*-

'''Manages accounts-related data in the user interface'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.px import Px
from appy.utils import formatNumber

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Entry:
    '''Abstract base class for any entry'''
    def __init__(self, account, label, style, sep, cellTag):
        # The outer account
        self.account = account
        # The (translated) label for the amount
        self.label = label
        # Some style may be applied to the label entry
        self.style = style
        # A separator line may de drawn just before rendering the row
        self.sep = sep
        # The tag to use for dumping a cell: td or th
        self.cellTag = cellTag

    def renderRow(self, i):
        '''Returns the table row corresponding to this entry, which has index
           p_i in the list.'''
        css = 'even' if i % 2 else 'odd'
        return '<tr class="%s">%s</tr>' % (css, self.render())

    def renderLabel(self):
        '''Renders the label for this entry'''
        s = self.style
        label = self.label if not s else '<%s>%s</%s>' % (s, self.label, s)
        return '<label>%s</label>' % label

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Amount(Entry):
    '''Represents one entry in an account, being some positive or negative
       amount.'''

    def __init__(self, account, label, value, positive=True, style=None,
                 unit=None, align='right', sep=False, cellTag='td'):
        # Call the base constructor
        Entry.__init__(self, account, label, style, sep, cellTag)
        # The (float or int) amount (or None)
        self.value = value
        # Must we negate the value ?
        self.positive = positive
        # The unit for this amount
        self.unit = unit or account.defaultUnit
        # Value alignment
        self.align = align

    def renderUnit(self, value, css=None):
        '''Render the "unit" part of this amount'''
        r = '' if value is None else (self.unit or '')
        if r:
            r = '<span class="discreet">%s</span>' % r
        t = self.cellTag
        return '<%s%s>%s</%s>' % (t, css or '', r, t)

    def render(self):
        '''Returns the XHTML table cells corresponding to this entry'''
        # Dump a separator if required
        tdc = self.sep and ' class="accountSep"' or ''
        # Format the value
        value = self.value
        if value is None:
            value = '-'
            sign = ''
        else:
            value = formatNumber(value, tsep=' ')
            sign = not self.positive and '-' or ''
        # Define the unit
        unit = self.renderUnit(value, css=tdc)
        # Produce the complete result
        t = self.cellTag
        return '<%s>%s</%s><%s%s align="center">%s</%s><%s align="%s"%s>%s' \
               '</%s>%s' % (t, self.renderLabel(), t, t, tdc, sign, t, t,
                            self.align, tdc, value, t, unit)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Section(Entry):
    '''Entry representing a section'''
    def __init__(self, account, label, style=None, sep=False, cellTag='td'):
        # Call the base constructor
        Entry.__init__(self, account, label, style, sep, cellTag)

    def render(self):
        '''Renders this section'''
        t = self.cellTag
        return '<%s class="accountSub">%s</%s><%s colspan="3"></%s>' % \
               (t, self.renderLabel(), t, t, t)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Account:
    '''An "account" is a list of entries, each of them being an amount (positive
       or negative), a total or some specific UI element like a separator.'''

    # Default width for columns: label, sign, value and unit
    defaultColumns = ['185em', '', '70em', '']

    def __init__(self, o, width='100%', css='', columns=None, defaultUnit=None):
        # The object from which data will be collected to create the account
        self.o = o
        # The account entries
        self.entries = []
        # Styles to apply to the resulting XHTML table
        self.width = width
        self.css = css
        self.columns = columns or Account.defaultColumns
        # This default unit, if present, will be shown for every entry for which
        # no unit is defined.
        self.defaultUnit = defaultUnit

    def addAmount(self, field=None, value=None, expression=None, label=None,
                  mapping=None, positive=True, style=None, unit=None,
                  sep=False, fieldValue=None, cellTag='td'):
        '''Adds an entry of type "amount" in the account'''

        # ----------------------------------------------------------------------
        # The label and value of the amount to display can be determined by:
        # ----------------------------------------------------------------------
        # p_field      | it must be the name of a Float or Integer field whose
        #              | value will be retrieved on p_self.o, excepted if
        #              | p_fieldValue is given: it will be used instead;
        # ----------------------------------------------------------------------
        # p_value      | it must be a float or int value. in that case, an i18n
        #              | p_label must be present and will be used (with an
        #              | optional p_mapping) to determine the label for this
        #              | amount;
        # ----------------------------------------------------------------------
        # p_expression | it must be a string containing a Python expression that
        #              | will be evaluated with "o" in its context ("o" being
        #              | p_self.o). The expression must return the float or int
        #              | amount; in that case, similarly to the previous case,
        #              | an i18n p_label must be present, with its optional
        #              | p_mapping.
        # ----------------------------------------------------------------------
        # p_positive is False, the amount will be negated. A specific p_style
        # can be applied to the entry "b" (bold) or "i" (italic).
        o = self.o
        if expression:
            label = o.translate(label, mapping=mapping)
            value = eval(expression)
        elif value is not None:
            label = o.translate(label, mapping=mapping)
        else:
            field = o.getField(field)
            label = o.translate(field.labelId)
            value = getattr(o, field.name) if fieldValue is None else fieldValue
        # Create the Amount instance and add it among this account's entries
        amount = Amount(self, label, value, positive=positive, style=style,
                        unit=unit, sep=sep, cellTag=cellTag)
        self.entries.append(amount)

    def addSection(self, label, style=None, sep=False, cellTag='td'):
        '''Adds a section entry'''
        section = Section(self, label, style=style, sep=sep, cellTag=cellTag)
        self.entries.append(section)

    def pop(self):
        '''Removes and return the last entry'''
        return self.entries.pop()

    def getColGroup(self):
        '''Get column specifiers for the main table'''
        cols = []
        for width in self.columns:
            cols.append('<col width="%s"/>' % width)
        return '<colgroup>%s</colgroup>' % ''.join(cols)

    px = Px('''
     <table with=":self.width" class=":self.css" id="account">
      <!-- Column specifiers -->
      <colgroup><col for="width in self.columns" width=":width"/></colgroup>
      <x for="entry in self.entries">::entry.renderRow(loop.entry.nb)</x>
     </table>''',

    css='''
     #account td { padding: 8px }
     .accountSep { border-top: 1px solid black }
     .accountSub { border-bottom: 1px dashed black }
     ''')

    def render(self):
        '''Renders the account as a chunk of XHTML'''
        return Account.px({'self': self})
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.ui.layout import Layout, Layouts
from appy.model.fields import Field

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Info(Field):
    '''An info is a field whose purpose is to present information
       (text, html...) to the user.'''

    class Layouts(Layouts):
        '''Info-specific layouts'''
        b = Layouts(edit='l')
        d = Layouts(edit=Layout('l-d', width=None))
        c = Layouts(edit='l|')
        dc = Layouts(edit='l|-d|')
        do = Layouts(edit='f', view='d') # Description only
        vdc = Layouts(edit='l', view='l|-d|')

        @classmethod
        def getDefault(class_, field):
            '''Default layouts for this Info p_field'''
            return class_.b

    # An info only displays a label. So PX for showing content are empty.
    view = edit = cell = search = ''

    def __init__(self, validator=None, multiplicity=(1,1), show='view',
      page='main', group=None, layouts=None, move=0, readPermission='read',
      writePermission='write', width=None, height=None, maxChars=None,
      colspan=1, master=None, masterValue=None, focus=False, historized=False,
      mapping=None, generateLabel=None, label=None, view=None, edit=None,
      cell=None, xml=None, translations=None):
        Field.__init__(self, None, (0,1), None, None, show, page, group,
          layouts, move, False, True, None, None, False, None, readPermission,
          writePermission, width, height, None, colspan, master, masterValue,
          focus, historized, mapping, generateLabel, label, None, None, None,
          None, False, False, view, cell, edit, xml, translations)
        self.validable = False
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import string
try:
    import ldap
except ImportError:
    ldap = None # For people that do not care about ldap

from appy.utils import sequenceTypes
from appy.server.cookie import Cookie
from appy.model.utils import Object as O

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
LDAP_DISABLED   = 'LDAP config not enabled.'
LDAP_CONNECT_KO = 'Could not connect to %s (%s).'
LDAP_COUNTS     = '%d local active SSO users // %d LDAP users'
LDAP_ACT        = '%d user(s) %sactivated: %s.'
LDAP_NO_DEACT   = 'No user was deactivated.'
SYNC_SUMMARY    = 'users synchronization: %d local user(s) created, %d ' \
                  'updated, %d untouched and %d invalid entries. %s local ' \
                  'user(s) deactivated and %d reactivated.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Config:
    '''Parameters for authenticating users to an LDAP server. This class is
       used by Appy apps. For a pure, appy-independent LDAP connector,
       see the class LdapConnector below.'''
    ldapAttributes = { 'loginAttribute': None, 'emailAttribute': 'email',
                       'fullNameAttribute': 'title',
                       'firstNameAttribute': 'firstName',
                       'lastNameAttribute': 'name' }
    # This config can be tested by trying to connect to the defined LDAP server
    # with the given "admin" credentials.
    testable = True

    def __init__(self):
        self.server = '' # Name of the LDAP server
        self.port = None # Port for this server
        self.protocol = 'ldap' # or ldaps
        # Login and password of the technical power user that the Appy
        # application will use to connect to the LDAP. Sometimes the full user
        # DN must be given.
        self.adminLogin = ''
        self.adminPassword = ''
        # LDAP attribute to use as login for authenticating users. Can also be:
        # "mail", "sAMAccountName", "cn", "nid", "uid"...
        self.loginAttribute = 'dn'
        # When a login is transmitted to the LDAP server, a transform can be
        # applied: upper, lower or capitalize.
        self.loginTransform = 'lower'
        # LDAP attributes for storing email
        self.emailAttribute = None
        # LDAP attribute for storing full name (first + last name)
        self.fullNameAttribute = None
        # Alternately, LDAP attributes for storing 1st & last names separately
        self.firstNameAttribute = None
        self.lastNameAttribute = None
        # Beyond the hereabove-defined attributes, you can specify here a
        # dict mapping additional LDAP attributes to attributes of the Appy User
        # class or its app-specific sub-class.
        self.customAttributes = {}
        # LDAP classes defining the users stored in the LDAP
        self.userClasses = ('top', 'person')
        # Based on "userClasses" above, a filter will be created to query users.
        # If you want to specify a more precise filter, specify it here. The
        # complete filter will be a logical AND between the base filter created
        # from "userClasses" and "userFilter". For example, to retrieve only
        # users having an email and those not being of a given class, you could
        # write something like:
        #         ldap.userFilter='(mail=*)(!(objectClass=Computer))'
        # (of course, the precise expression depends on the attributes and
        #  values as defined in your LDAP).
        self.userFilter = ''
        # Base DN where to find users in the LDAP. Several base DNs can be
        # specified, in a list or tuple.
        self.baseDn = ''
        self.scope = 'SUBTREE' # Scope of the search within self.baseDn
        # Is this server connection enabled ?
        self.enabled = True
        # The "user map" allows to put LDAP users into groups or assign them
        # roles. This dict will be used every time a local User will be created.
        # It can be while synchronizing all users (see m_synchronizeUsers
        # below) or when the user logs in for the first time (see m_getUser
        # below). This dict will NOT be used subsequently, when updating the
        # User instance. Every key must be a user login. Every value is an
        # appy.Object instance having the optional attributes:
        # "groups": a list of group IDs (logins);
        # "roles":  a list of global role names.
        self.userMap = {}
        # If the following attribute is True, when successfully authenticating a
        # user from this LDAP, if the corresponding User instance is a local
        # one, it will be converted to a distant copy, excepted if it is a
        # special user.
        self.convertLocalUsers = True
        # While synchronizing distant users with their local copies, the
        # following attribute determines if locally deactivated users being
        # found in the LDAP are reactivated or not. By default, they are
        # reactivated, the distant source being considered as the "authentic
        # source of data" for its users.
        self.reactivateUsers = True

    def init(self, tool):
        '''Lazy initialization'''
        # Standardize attributes
        if not isinstance(self.baseDn, sequenceTypes):
            self.baseDn = [self.baseDn]

    def test(self, tool):
        '''Connects to the LDAP server to check the connection'''
        connector = LdapConnector(self.getServerUri(), tool=tool)
        success, msg = connector.connect(self.adminLogin, self.adminPassword)
        return success and 'OK' or 'KO'

    def __repr__(self):
        '''Short string representation of this ldap config, for logging and
           debugging purposes.'''
        return 'ldap: %s' % self.getServerUri()

    def getServerUri(self):
        '''Returns the complete URI for accessing the LDAP, ie
           "ldap[s]://some.ldap.server:389".'''
        port = self.port or 389
        return '%s://%s:%d' % (self.protocol, self.server, port)

    def getUserFilterValues(self, login=None):
        '''Gets the filter values required to perform a query for finding user
           corresponding to p_login in the LDAP, or all users if p_login is
           None.'''
        res = login and [(self.loginAttribute, login)] or []
        for userClass in self.userClasses:
            res.append( ('objectClass', userClass) )
        return res

    def getUserAttributes(self):
        '''Gets the attributes we want to get from the LDAP for characterizing
           a user.'''
        r = []
        for name in self.ldapAttributes:
            attrName = getattr(self, name)
            if attrName: r.append(attrName)
        # Add custom attributes
        for name in self.customAttributes:
            if name not in r: r.append(name)
        return r

    def setAppyValue(self, ldapName, ldapData, appyName, appyData):
        '''Convert LDAP value found at key p_ldapName in p_ldapData to an
           Appy-compliant value and store it, if found, in p_appyData at key
           p_appyName.'''
        value = ldapData.get(ldapName)
        if not value: return
        # Convert the value when relevant
        if isinstance(value, list): value = value[0]
        appyData[appyName] = value.decode('utf-8')

    def getUserParams(self, ldapData):
        '''Formats the user-related p_ldapData retrieved from the LDAP, as a
           dict of params usable for creating or updating the corresponding
           Appy user.'''
        res = {}
        # Retrieve base params
        for name, appyName in self.ldapAttributes.items():
            if not appyName: continue
            # Get the name of the attribute as known in the LDAP
            ldapName = getattr(self, name)
            if not ldapName: continue
            self.setAppyValue(ldapName, ldapData, appyName, res)
        # Retrieve custom params
        for ldapName, appyName in self.customAttributes.items():
            if ldapData.has_key(ldapName):
                self.setAppyValue(ldapName, ldapData, appyName, res)
        return res

    def setLocalUser(self, tool, attrs, login, password=None, sso=None):
        '''Creates or updates the local User instance corresponding to a LDAP
           user from the LDAP, having p_login. Its other attributes are in
           p_attrs and, when relevant, its password is in p_password. This
           method r_eturns a 2-tuple containing:
           * the local User instance;
           * the status of the operation:
             - "created" if the instance has been created,
             - "updated" if at least one data from p_attrs is different from the
               one stored on the existing User instance;
             - None else.
        '''
        # This requires a commit
        handler = tool.H()
        handler.commit = True
        # The LDAP config may come from a SSO config
        source = 'sso' if sso else 'ldap'
        # Do we already have a local User instance for this user ?
        status = None
        user = tool.search1('User', login=login)
        if user: # yes
            # Do not update it if the user is special
            if user.isSpecial(): return None, None
            # If the user is local, don't touch it or convert it to LDAP
            if user.source == 'zodb':
                if self.convertLocalUsers:
                    # Convert it to a LDAP (or SSO) user
                    user.convertTo(source)
                else:
                    return None, None
            # Update it with info about him from the LDAP
            for name, value in attrs.items():
                currentValue = getattr(user, name)
                if value != currentValue:
                    setattr(user, name, value)
                    status = 'updated'
            # Recompute the title, computed from other fields
            user.updateTitle()
            user.reindex()
        else:
            # Create the user
            user = tool.create('users', login=login, source=source, **attrs)
            status = 'created'
            # Put him into groups and/or grant him some roles according to
            # self.userMap.
            if login in self.userMap:
                privileges = self.userMap[login]
                # Put the user in some groups
                groups = getattr(privileges, 'groups', None)
                if groups:
                    for groupLogin in groups:
                        group = tool.search1('Group', login=groupLogin)
                        group.link('users', user)
                # Grant him some roles
                roles = getattr(privileges, 'roles', None)
                if roles:
                    for role in roles: user.addRole(role)
                tool.log('%s: automatic privileges set.' % login)
        # Update user password, if given
        if password:
            user.password = password
            Cookie.updatePassword(handler, password)
        return user, status

    def applyLoginTransform(self, login):
        '''Applies, when relevant, a transform to p_login'''
        if self.loginTransform:
            r = eval('login.%s()' % self.loginTransform)
        else:
            r = login
        return r

    def getUser(self, tool, login, password):
        '''Returns a local User instance corresponding to a LDAP user if p_login
           and p_password correspond to a valid LDAP user.'''
        # Check if LDAP is enabled
        if not self.enabled: return
        # Get a connector to the LDAP server and connect to the LDAP server
        serverUri = self.getServerUri()
        connector = LdapConnector(serverUri, tool=tool)
        success, msg = connector.connect(self.adminLogin, self.adminPassword)
        if not success: return
        # Apply a transform to p_login when required
        if login: login = self.applyLoginTransform(login)
        # Check if the user corresponding to p_login exists in the LDAP
        filter = connector.getFilter(self.getUserFilterValues(login),
                                     self.userFilter)
        params = self.getUserAttributes()
        data = None
        for baseDn in self.baseDn:
            data = connector.search(baseDn, self.scope, filter, params)
            if data: break
        if not data or not data[0][0]: return
        # The user exists. Try to connect to the LDAP with this user in order
        # to validate its password.
        userConnector = LdapConnector(serverUri, tool=tool)
        success, msg = userConnector.connect(data[0][0], password)
        if not success: return
        # The password is correct. We can create/update our local user
        # corresponding to this LDAP user.
        userParams = self.getUserParams(data[0][1])
        user, status = self.setLocalUser(tool, userParams, login, password)
        return user

    def updateUserStates(self, tool, ldapUsers, counts, sso=None):
        '''Browse local users and activate or deactivate some of them, based on
           freshly queried LDAP users from p_ldapUsers. Update p_counts of users
           (attributes "deactivated" and "reactivated").'''
        # Remember logins of (de)(re)activated users, for logging them
        deactivated = []
        reactivated = []
        source = 'sso' if sso else 'ldap'
        # Count the active local copies of external users
        local = 0
        # Must deactivated users be reactivated if found among p_ldapUsers ?
        mustReactivate = self.reactivateUsers
        # Browse local users
        for user in tool.users:
            # Ignore those not being external
            if user.source != source: continue
            # Deactivate appropriate users
            if user.state == 'active':
                local += 1
                if user.login not in ldapUsers:
                    deactivated.append('%s (%s)' % (user.login,user.getTitle()))
                    user.do('deactivate')
                    counts.deactivated += 1
            # Reactivate appropriate users
            elif mustReactivate and (user.state == 'inactive'):
                if user.login in ldapUsers:
                    reactivated.append('%s (%s)' % (user.login,user.getTitle()))
                    user.do('reactivate')
                    counts.reactivated += 1
        log = tool.log
        log(LDAP_COUNTS % (local, len(ldapUsers)))
        if deactivated:
            log(LDAP_ACT % (counts.deactivated, 'de', ', '.join(deactivated)))
        elif mustReactivate:
            log(LDAP_NO_DEACT)
        if reactivated:
            log(LDAP_ACT % (counts.reactivated, 're', ', '.join(reactivated)))

    def synchronizeUsers(self, tool, sso=None):
        '''Synchronizes the local User copies with this LDAP user base. Returns
           a message with details about the operation.'''
        # If this LDAP config is part of a SSO config, this latter is available
        # in p_sso ; in that case, users need to be flagged as having source
        # "sso" and not "ldap".
        if not self.enabled: raise Exception(LDAP_DISABLED)
        # Get a connector to the LDAP server and connect to the LDAP server
        serverUri = self.getServerUri()
        tool.log('reading users from %s...' % serverUri)
        connector = LdapConnector(serverUri, tool=tool)
        success, msg = connector.connect(self.adminLogin, self.adminPassword)
        if not success: raise Exception(LDAP_CONNECT_KO % (serverUri, msg))
        # Query the LDAP for users. Perform several queries to avoid having
        # error ldap.SIZELIMIT_EXCEEDED.
        params = self.getUserAttributes()
        # Initialise user counts
        counts = O(created=0, updated=0, untouched=0, wrong=0,
                   deactivated=0, reactivated=0)
        # Remember the logins of encountered LDAP users: it is required to
        # (de)activate local users.
        ldapUsers = {} # ~{s_login: None}~
        firstChars = string.ascii_lowercase + string.digits
        for char in firstChars:
            # Get all the users whose login starts with "char"
            filter = connector.getFilter(self.getUserFilterValues('%s*' % char),
                                         self.userFilter)
            for baseDn in self.baseDn:
                data = connector.search(baseDn, self.scope, filter, params)
                if not data: continue
                for userData in data:
                    # Get the user login
                    try:
                        login = userData[1][self.loginAttribute][0]
                        login = self.applyLoginTransform(login.decode('utf-8'))
                    except TypeError:
                        # The retrieved data structure is not valid
                        counts.wrong += 1
                        continue
                    # Get the other user parameters, as Appy wants it
                    userParams = self.getUserParams(userData[1])
                    # Create or update the user
                    user, status = self.setLocalUser(tool, userParams, login,
                                                     sso=sso)
                    if status == 'created': counts.created += 1
                    elif status == 'updated': counts.updated += 1
                    else: counts.untouched += 1
                    if user: ldapUsers[login] = None
        # Activate or deactivate users
        self.updateUserStates(tool, ldapUsers, counts, sso=sso)
        # Log the operation and return a message
        message = SYNC_SUMMARY % (counts.created, counts.updated,
                                  counts.untouched, counts.wrong,
                                  counts.deactivated, counts.reactivated) 
        tool.log(message)
        return message

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class LdapConnector:
    '''This class manages the communication with a LDAP server'''
    def __init__(self, serverUri, tentatives=5, ssl=False, timeout=5,
                 tool=None):
        # The URI of the LDAP server, ie ldap[s]://some.ldap.server:389.
        self.serverUri = serverUri
        # The object that will represent the LDAP server
        self.server = None
        # The number of trials the connector will at most perform to the LDAP
        # server, when executing a query in it.
        self.tentatives = tentatives
        self.ssl = ssl
        # The timeout for every query to the LDAP.
        self.timeout = timeout
        # A tool from a Appy application can be given and will be used, ie for
        # logging purpose.
        self.tool = tool

    def log(self, message, type='info'):
        '''Logs via a Appy tool if available'''
        if self.tool:
            self.tool.log(message, type=type)
        else:
            print(message)

    def connect(self, login, password):
        '''Connects to the LDAP server using p_login and p_password as
           credentials. If the connection succeeds, a server object is created
           in self.server and tuple (True, None) is returned. Else, tuple
           (False, errorMessage) is returned.'''
        try:
            self.server = ldap.initialize(self.serverUri)
            self.server.simple_bind_s(login, password)
            # Tentative code
            #self.server.protocol_version = 3
            #self.server.set_option(ldap.OPT_REFERRALS, 0)
            # Note that "simple_bind_s" can return a result
            return True, None
        except AttributeError as ae:
            # When the ldap module is not there, trying to catch ldap.LDAPError
            # will raise an error.
            message = str(ae)
            self.log('Ldap connect error with login %s (%s).' % \
                     (login, message))
            return False, message
        except ldap.LDAPError as le:
            message = str(le)
            self.log('%s: connect error with login %s (%s).' % \
                     (self.serverUri, login, message))
            return False, message

    def getFilter(self, values, customPart=''):
        '''Builds and returns a LDAP filter based on p_values, a tuple of
           (name, value) expressions logically AND-ed. If p_customPart is given,
           it represents and additional expression that will be AND-ed as
           well.'''
        return '(&%s%s)' % (''.join(['(%s=%s)' % (n, v) for n, v in values]),
                            customPart)

    def search(self, baseDn, scope, filter, attributes=None):
        '''Performs a query in the LDAP at node p_baseDn, with the given
           p_scope. p_filter is a LDAP filter that constraints the search. It
           can be computed from a list of tuples (value, name) by method
           m_getFilter. p_attributes is the list of attributes that we will
           retrieve from the LDAP. If None, all attributes will be retrieved.'''
        if self.ssl: self.server.start_tls_s()
        try:
            # Get the LDAP constant corresponding to p_scope
            scope = getattr(ldap, 'SCOPE_%s' % scope)
            # Perform the query
            for i in range(self.tentatives):
                try:
                    return self.server.search_st(\
                        baseDn, scope, filterstr=filter, attrlist=attributes,
                        timeout=self.timeout)
                except ldap.TIMEOUT:
                    pass
        except ldap.LDAPError as le:
            self.log('LDAP query error %s: %s' % \
                     (le.__class__.__name__, str(le)))
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.model.base import Base
from appy.model.fields import Show
from appy.ui.layout import Layouts
from appy.xml.escape import Escape
from appy.model.fields.file import File
from appy.model.fields.list import List
from appy.model.fields.color import Color
from appy.model.fields.action import Action
from appy.model.fields.string import String
from appy.model.fields.select import Select, Selection
from appy.model.workflow.standard import TooPermissive

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Document(Base):
    '''Base class representing a binary document: image, file, etc.'''

    # Use the TooPermissive standard workflow, complemented by methods
    # m_mayView, m_mayEdit and m_mayDelete defined hereafter.
    workflow = TooPermissive

    # Documents are not indexed by default
    indexable = False
    popup = ('500px', '500px')
    listColumns = ('thumb*60px|', 'title')

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #                               The file
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    do = {'label': 'Document'}

    def getMaxWidth(self):
        '''Get the maximum width for uploaded images. If the image is larger
           than this, it will be resized.'''
        # The size depends on the container
        try:
            container = self.container or self.initiator.o
        except AttributeError:
            # There is no container yet if the image is uploaded via ckeditor
            container = self.traversal.o
        return '%dpx' % container.maxWidth

    def getViewWidth(self):
        '''The width of the image as shown on the "view" layout varies depending
           on the it to be rendered in the popup or no.'''
        return '450px' if self.req.popup == 'True' else '700px'

    file = File(multiplicity=(1,1), isImage=True, resize=True, cache=True,
                width=getMaxWidth, viewWidth=getViewWidth, nameStorer='title',
                thumbnail='thumb', **do)

    # Its thumbnail
    thumb = File(isImage=True, resize=True, width='100px', show=Show.VE_, **do)

    def getCarousel(self):
        '''Returns the carousel into which this document is included'''
        container = self.container
        if container and (container.class_.name == 'Carousel'):
            return container

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #                  Elements to render on top of the file
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # When used in a carousel, elements may be specified, that must be shown
    # on top of the document.
    elementTypes = ('title', 'underTitle', 'buttonA', 'buttonB')

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #                          Action "duplicate"
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # This action creates a clone B from a document A, and adds B to A's
    # container, at the end of the same Ref linking the container to A.

    def showDuplicate(self):
        '''Shows this action only to the user being allowed to edit the
           document's container.'''
        carousel = self.getCarousel()
        return 'buttons' if carousel and carousel.allows('write') else None

    def doDuplicate(self):
        '''Creates a clone from p_self and links it to p_self's container'''
        container, name = self.getContainer(forward=True)
        clone = container.createFrom(name, self)
        self.say(self.translate('action_done'))
        return True, self.referer

    duplicate = Action(show=showDuplicate, action=doDuplicate,
                       result='redirect', confirm=True, **do)

    @classmethod
    def listElementTypes(class_, o):
        '''Lists the possible types for elements' entries'''
        return [(type, o.translate('et_%s' % type)) \
                for type in class_.elementTypes]

    sub = (
      ('type', Select(validator=Selection(lambda o: o.listElementTypes(o)),
                      multiplicity=(1,1))),
      ('text', String(multiplicity=(1,1))),
      ('position', String(multiplicity=(1,1), width=6, default='40% 5%')),
      ('url', String()) # No URL validator: JS code is allowed (javascript:...)
    )
    elements = List(sub, show=lambda o: o.container.class_.name == 'Carousel',
                    layouts=Layouts.td, **do)

    def getElementLink(self, carousel, element, text):
        '''Returns the compete "a" tag for a clickable element'''
        # Get link attributes from the p_carousel, if any
        attributes = carousel.linkAttributes
        target = onClick = params = None
        if attributes and (element.type in attributes):
            attrs = attributes[element.type]
            target = attrs.target
            onClick = attrs.onClick
            params = attrs.params
        target = (' target="%s"' % target) if target else ''
        onClick = (' onClick="%s"' % onClick) if onClick else ''
        url = element.url
        if params:
            sep = '&' if '?' in url else '?'
            url += sep + params
        return '<a href="%s"%s%s>%s</a>' % (url, target, onClick, text)

    def getElements(self, carousel):
        '''Returns, as a list of chunks of XHTML code, the list of elements from
           field "elements", ready to be dumped on top of p_self in this
           p_carousel.'''
        r = []
        if self.isEmpty('elements'): return r
        for element in self.elements:
            # Compute the element's position, relative to s_self's top left
            # corner.
            top, left = element.position.split(' ')
            # Get the CSS class to apply
            css = carousel.getCssName(element.type)
            # Get the text to display
            text = Escape.xhtml(element.text)
            if element.url:
                # Wrap the text in a clickable link
                text = self.getElementLink(carousel, element, text)
            # Produce the complete chunk of XHTML for this element
            html = '<div class="%s" style="top:%s;left:%s">%s</div>' % \
                   (css, top, left, text)
            r.append(html)
        return r

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #                            Main methods
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def validate(self, new, errors):
        '''Validate elements'''
        elements = new.elements
        if not elements: return
        i = -1
        for element in elements:
            i += 1
            # Validate the position
            parts = element.position.split(' ')
            if len(parts) != 2:
                message = self.translate('position_ko')
                setattr(errors, 'elements*position*%d' % i, message)

    def mayView(self):
        '''This document is viewable if its container is'''
        return self.container.allows('read')

    def mayEdit(self):
        '''This document is editable if its container is'''
        # The container is not present just before the newly created document is
        # tied to its initiator.
        container = self.container
        return not container or container.allows('write')

    def mayDelete(self):
        '''This document can be deleted if its container can be deleted, too'''
        return self.container.allows('delete')
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import time

from DateTime import DateTime
from DateTime.interfaces import DateError

from appy.px import Px
from appy.utils import dates
from appy.model.fields import Field
from appy.model.fields.hour import Hour
from appy.database.operators import in_
from appy.database.indexes.date import DateIndex

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Date(Field):

    # Required CSS and Javascript files for this type
    cssFiles = {'edit': ('jscalendar/calendar-blue.css',)}
    jsFiles = {'edit': ('jscalendar/calendar.js',
                        'jscalendar/lang/calendar-en.js',
                        'jscalendar/calendar-setup.js')}

    # Possible values for "format"
    WITH_HOUR = 0
    WITHOUT_HOUR = 1
    dateParts = ('year', 'month', 'day')
    hourParts = ('hour', 'minute')

    # Default value on the search screen
    searchDefault = (None, None, None)

    # Precision of the indexed value, in minutes
    indexPrecision = 1

    # The name of the index class storing values of this field in the catalog
    indexType = 'DateIndex'

    view = cell = Px('''<x>:value</x>''')

    # PX for selecting hour and minutes, kindly provided by field Hour
    pxHour = Hour.edit

    edit = Px('''
     <x var="years=field.getSelectableYears(o)">
      <!-- Day -->
      <select if="field.showDay" var2="days=range(1,32); part='%s_day' % name"
              name=":part" id=":part">
       <option value="">-</option>
       <option for="day in days" var2="zDay=str(day).zfill(2)" value=":zDay"
         selected=":field.isSelected(o, part, 'day', \
                                     day, rawValue)">:zDay</option>
      </select> 

      <!-- Month -->
      <select var="months=range(1,13); part='%s_month' % name"
              name=":part" id=":part">
       <option value="">-</option>
       <option for="month in months"
         var2="zMonth=str(month).zfill(2)" value=":zMonth"
         selected=":field.isSelected(o, part, 'month', \
                                     month, rawValue)">:zMonth</option>
      </select> 

      <!-- Year -->
      <select var="part='%s_year' % name" name=":part" id=":part">
       <option value="">-</option>
       <option for="year in years" value=":year"
         selected=":field.isSelected(o, part, 'year', \
                                     year, rawValue)">:year</option>
      </select>

      <!-- The icon for displaying the calendar popup -->
      <x if="field.calendar">
       <input type="hidden" id=":name" name=":name"/>
       <img id=":'%s_img' % name" src=":url('calendar.svg')" class="iconS"/>
       <script>::field.getJsInit(name, years)</script>
      </x>

      <!-- Hour and minutes -->
      <x if="field.format == 0">:field.pxHour</x>
     </x>''')

    search = Px('''
     <table var="years=field.getSelectableYears(o);
                 dstart,dend=field.getDefaultSearchValues(o)">
       <!-- From -->
       <tr var="fromName='%s_from' % name;
                dayFromName='%s_from_day' % name;
                monthFromName='%s_from_month' % name">
        <td width="10px">&nbsp;</td>
        <td><label>:_('search_from')</label></td>
        <td>
         <select id=":dayFromName" name=":dayFromName">
          <option value="">--</option>
          <option for="value in [str(v).zfill(2) for v in range(1, 32)]"
                  value=":value" selected=":value == dstart[2]">:value</option>
         </select> / 
         <select id=":monthFromName" name=":monthFromName">
          <option value="">--</option>
          <option for="value in [str(v).zfill(2) for v in range(1, 13)]"
                  value=":value" selected=":value == dstart[1]">:value</option>
         </select> / 
         <select id=":widgetName" name=":widgetName">
          <option value="">--</option>
          <option for="value in years"
                  value=":value" selected=":value == dstart[0]">:value</option>
         </select>
         <!-- The icon for displaying the calendar popup -->
         <x if="field.calendar">
          <input type="hidden" id=":fromName" name=":fromName"/>
          <img id=":'%s_img' % fromName" src=":url('calendar.svg')"
               class="iconS"/>
          <script>::field.getJsInit(fromName, years)</script>
         </x>
         <!-- Hour and minutes when relevant -->
         <x if="(field.format == 0) and field.searchHour"
            var2="hPart='%s_from_hour' % name;
                  mPart='%s_from_minute' % name">:field.pxHour</x>
        </td>
       </tr>

       <!-- To -->
       <tr var="toName='%s_to' % name;
                dayToName='%s_to_day' % name;
                monthToName='%s_to_month' % name;
                yearToName='%s_to_year' % name">
        <td></td>
        <td><label>:_('search_to')</label>&nbsp;&nbsp;&nbsp;&nbsp;</td>
        <td height="20px">
         <select id=":dayToName" name=":dayToName">
          <option value="">--</option>
          <option for="value in [str(v).zfill(2) for v in range(1, 32)]"
                  value=":value" selected=":value == dend[2]">:value</option>
         </select> / 
         <select id=":monthToName" name=":monthToName">
          <option value="">--</option>
          <option for="value in [str(v).zfill(2) for v in range(1, 13)]"
                  value=":value" selected=":value == dend[1]">:value</option>
         </select> / 
         <select id=":yearToName" name=":yearToName">
          <option value="">--</option>
          <option for="value in years"
                  value=":value" selected=":value == dend[0]">:value</option>
         </select>
         <!-- The icon for displaying the calendar popup -->
         <x if="field.calendar">
          <input type="hidden" id=":toName" name=":toName"/>
          <img id=":'%s_img' % toName" src=":url('calendar.svg')"
               class="iconS"/>
          <script>::field.getJsInit(toName, years)</script>
         </x>
         <!-- Hour and minutes when relevant -->
         <x if="(field.format == 0) and field.searchHour"
            var2="hPart='%s_to_hour' % name;
                  mPart='%s_to_minute' % name">:field.pxHour</x>
        </td>
       </tr>
      </table>''')

    def __init__(self, validator=None, multiplicity=(0,1), default=None,
      defaultOnEdit=None, format=WITH_HOUR, dateFormat=None, hourFormat=None,
      calendar=True, startYear=time.localtime()[0]-10,
      endYear=time.localtime()[0]+10, reverseYears=False, minutesPrecision=5,
      show=True, page='main', group=None, layouts=None, move=0, indexed=False,
      mustIndex=True, indexValue=None, emptyIndexValue=0, searchable=False,
      filterField=None, readPermission='read', writePermission='write',
      width=None, height=None, maxChars=None, colspan=1, master=None,
      masterValue=None, focus=False, historized=False, mapping=None,
      generateLabel=None, label=None, sdefault=None, scolspan=1, swidth=None,
      sheight=None, persist=True, view=None, cell=None, edit=None, xml=None,
      translations=None, showDay=True, searchHour=False):
        self.format = format
        self.calendar = calendar
        self.startYear = startYear
        self.endYear = endYear
        # If reverseYears is True, in the selection box, available years, from
        # self.startYear to self.endYear will be listed in reverse order.
        self.reverseYears = reverseYears
        # If p_showDay is False, the list for choosing a day will be hidden
        self.showDay = showDay
        # If no p_dateFormat/p_hourFormat is specified, the application-wide
        # tool.dateFormat/tool.hourFormat is used instead.
        self.dateFormat = dateFormat
        self.hourFormat = hourFormat
        # In the context of a Date, the max hour is always 23. But it can be
        # more in the context of an Hour field.
        self.maxHour = 23
        # If "minutesPrecision" is 5, only a multiple of 5 can be encoded. If
        # you want to let users choose any number from 0 to 59, set it to 1.
        self.minutesPrecision = minutesPrecision
        # The search widget will only allow to specify start and end dates
        # without hour, event if format is WITH_HOUR, excepted if searchHour is
        # True.
        self.searchHour = searchHour
        # Value for p_sdefault must be a tuple (start, end), where each value
        # ("start" or "end"), if not None, must be a tuple (year, month, day).
        # Each of these sub-values can be None or an integer value. p_sdefault
        # can also be a method that produces the values in the specified format.
        Field.__init__(self, validator, multiplicity, default, defaultOnEdit,
          show, page, group, layouts, move, indexed, mustIndex, indexValue,
          emptyIndexValue, searchable, filterField, readPermission,
          writePermission, width, height, None, colspan, master, masterValue,
          focus, historized, mapping, generateLabel, label, sdefault, scolspan,
          swidth, sheight, persist, False, view, cell, edit, xml, translations)

    def getCss(self, layout, r):
        '''CSS files are only required if the calendar must be shown'''
        if self.calendar: Field.getCss(self, layout, r)

    def getJs(self, layout, r, config):
        '''Javascript files are only required if the calendar must be shown'''
        if self.calendar: Field.getJs(self, layout, r, config)

    def getSelectableYears(self, o):
        '''Gets the list of years one may select for this field'''
        startYear = self.getAttribute(o, 'startYear')
        r = range(startYear, self.endYear + 1)
        if self.reverseYears: r.reverse()
        return r

    def validateValue(self, obj, value):
        try:
            value = DateTime(value)
        except (DateError, ValueError):
            return obj.translate('bad_date')

    def getFormattedValue(self, o, value, layout='view', showChanges=False,
                          language=None):
        if self.isEmptyValue(o, value): return ''
        # Get the applicable date format
        ui = o.config.ui
        dateFormat = self.dateFormat or ui.dateFormat
        # A problem may occur with some extreme year values. Replace the "year"
        # part "by hand".
        if '%Y' in dateFormat:
            dateFormat = dateFormat.replace('%Y', str(value.year()))
        r = ui.formatDate(o.tool, value, dateFormat, withHour=False)
        if self.format == Date.WITH_HOUR:
            r += ' %s' % value.strftime(self.hourFormat or ui.hourFormat)
        return r

    def getRequestValue(self, o, requestName=None):
        req = o.req
        name = requestName or self.name
        # Manage the "date" part
        value = ''
        for part in self.dateParts:
            # The "day" part may be hidden. Use "1" by default.
            if (part == 'day') and not self.showDay:
                valuePart = '01'
            else:
                valuePart = req['%s_%s' % (name, part)]
            if not valuePart: return
            value += valuePart + '/'
        value = value[:-1]
        # Manage the "hour" part
        if self.format == self.WITH_HOUR:
            value += ' '
            for part in self.hourParts:
                valuePart = req['%s_%s' % (name, part)]
                if not valuePart: return
                value += valuePart + ':'
            value = value[:-1]
        return value

    def searchValueIsEmpty(self, req):
        '''We consider a search value being empty if both "from" and "to" values
           are empty. At an individual level, a "from" or "to" value is
           considered not empty if at least the year is specified.'''
        # The base method determines if the "from" year is empty
        isEmpty = Field.searchValueIsEmpty
        return isEmpty(self, req) and \
               isEmpty(self, req, widgetName='%s_to_year' % self.name)

    def getRequestSuffix(self): return '_year'

    def getStorableValue(self, o, value, complete=False):
        if not self.isEmptyValue(o, value):
            return DateTime(value)

    def getDateFromSearchValue(self, year, month, day, hour, setMin):
        '''Gets the index representation of a valid DateTime instance, built
           from date information coming from the request.'''
        # This info comes as strings in p_year, p_month, p_day and p_hour.
        # The method returns None if p_year is empty. If p_setMin is True, when
        # some information is missing (month or day), it is replaced with the
        # minimum value (=1). Else, it is replaced with the maximum value
        # (=12, =31).
        if not year: return
        # Set month and day
        if not month:
            month = 1 if setMin else 12
        if not day:
            day = 1 if setMin else 31
        # Set the hour
        if hour is None:
            hour = '00:00' if setMin else '23:59'
        # The specified date may be invalid (ie, 2018/02/31): ensure to produce
        # a valid date in all cases.
        try:
            r = DateTime('%s/%s/%s %s' % (year, month, day, hour))
        except:
            base = DateTime('%s/%s/01' % (year, month))
            r = dates.getLastDayOfMonth(base, hour=hour)
        return r

    def getSearchPart(self, req, to=False, value=None):
        '''Gets the search value from p_req corresponding to the "from" or "to"
           part (depending on boolean p_to).'''
        name = self.name
        part = 'to' if to else 'from'
        # Get the year. For the "from" part, it corresponds to the name of
        # field. For the "to" part, there is a specific key in the request.
        year = req['%s_to_year' % name] \
               if to else Field.getSearchValue(self, req, value=value)
        month = req['%s_%s_month' % (name, part)]
        day = req['%s_%s_day' % (name, part)]
        hour = None
        if self.searchHour:
            hour = '%s:%s' % (req['%s_%s_hour' % (name, part)] or '00',
                              req['%s_%s_minute' % (name, part)] or '00')
        return self.getDateFromSearchValue(year, month, day, hour, not to)

    def getSearchValue(self, req, value=None):
        '''Converts raw search values from p_req into an interval of dates'''
        return in_(self.getSearchPart(req, False, value),
                   self.getSearchPart(req, True, value))

    def getDefaultSearchValues(self, o):
        '''Gets the default value for this field when shown on a search
           layout.'''
        default = self.getAttribute(o, 'sdefault')
        if not default:
            r = self.searchDefault, self.searchDefault
        else:
            # Convert months and days to zfilled strings
            r = []
            for i in (0, 1):
                value = default[i]
                if value:
                    year, month, day = value
                    if month is not None: month = str(month).zfill(2)
                    if day is not None: day = str(day).zfill(2)
                    value = year, month, day
                else:
                    value = self.searchDefault
                r.append(value)
        return r

    def isSelected(self, o, part, fieldPart, dateValue, dbValue):
        '''When displaying this field, must the particular p_dateValue be
           selected in the sub-field p_fieldPart corresponding to the date
           part ?'''
        # Get the value we must compare (from request or from database)
        req = o.req
        if part in req:
            compValue = req[part]
            if compValue.isdigit():
                compValue = int(compValue)
        else:
            compValue = dbValue
            if compValue:
                compValue = getattr(compValue, fieldPart)()
        # Compare the value
        return compValue == dateValue

    def isSortable(self, inRefs=False):
        '''Can this field be sortable ?'''
        return True if inRefs else Field.isSortable(self) # Sortable in Refs

    def getJsInit(self, name, years):
        '''Gets the Javascript init code for displaying a calendar popup for
           this field, for an input named p_name (which can be different from
           self.name if, ie, it is a search field).'''
        # Always express the range of years in chronological order.
        years = [years[0], years[-1]]
        years.sort()
        return 'Calendar.setup({inputField: "%s", button: "%s_img", ' \
               'onSelect: onSelectDate, range:%s, firstDay: 1})' % \
               (name, name, str(years))
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.xml import Environment, Parser
from appy.pod.buffers import MemoryBuffer
from appy.pod.xhtml.tags import SELF_CLOSING

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class PxEnvironment(Environment):
    '''Environment for the PX parser'''

    def __init__(self):
        # In the following buffer, we will create a single memory sub-buffer
        # that will hold the result of parsing the PX = a hierarchy of memory
        # buffers = PX's AST (Abstract Syntax Tree).
        # A major difference between POD and PX: POD creates the AST and
        # generates the result in the same step: one AST is generated, and then
        # directly produces a single evaluation, in the root file buffer. PX
        # works in 2 steps: the AST is initially created in self.ast. Then,
        # several (concurrent) evaluations can occur, without re-generating the
        # AST.
        self.ast = MemoryBuffer(self, None)
        # Buffer where we must dump the content we are currently reading
        self.currentBuffer = self.ast
        # Tag content we are currently reading. We will put something in this
        # attribute only if we encounter content that is Python code.
        # Else, we will directly dump the parsed content into the current
        # buffer.
        self.currentContent = ''
        # The currently walked tag. We redefine it here. This attribute is
        # normally managed by the parent Environment, but we do not use the
        # standard machinery from this environmment and from the default Parser
        # for better performance. Indeed, the base parser and env process
        # namespaces, and we do not need this for the PX parser.
        self.currentTag = None
        # Exceptions are always raised (for pod, it is not the case)
        self.raiseOnError = True

    def addSubBuffer(self):
        subBuffer = self.currentBuffer.addSubBuffer()
        self.currentBuffer = subBuffer

    def isActionElem(self, elem):
        '''Returns True if the currently walked p_elem is the same elem as the
           main buffer elem.'''
        action = self.currentBuffer.action
        return action and (action.elem == elem)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class PxParser(Parser):
    '''PX parser that is specific for parsing PX data'''
    pxAttributes = ('var', 'for', 'if', 'var2')
    noDumpTags = ('selected', 'checked', 'disabled', 'multiple', 'readonly')
    # The following dict allows to convert attrs "lfor" to "for". Indeed,
    # because tags "label" can have an attribute named "for", it clashes with
    # the "for" attribute added by PX. The solution is to force users to write,
    # in their PX, the HTML attr for" as "lfor".
    renamedAttributes = {'lfor': 'for'}

    def __init__(self, env, caller=None):
        Parser.__init__(self, env, caller)

    def startElement(self, elem, attrs):
        '''A start p_elem with p_attrs is encountered in the PX.'''
        e = self.env
        self.currentTag = elem
        # See if we have a PX attribute among p_attrs
        found = False
        for name in self.pxAttributes:
            if name in attrs:
                if not found:
                    # This is the first PX attr we find.
                    # Create a sub-buffer with an action.
                    e.addSubBuffer()
                    found = True
                # Add the action
                e.currentBuffer.createPxAction(elem, name, attrs[name])
        if e.isActionElem(elem):
            # Add a temp element in the buffer (that will be unreferenced
            # later). This way, when encountering the corresponding end element,
            # we will be able to check whether the end element corresponds to
            # the main element or to a sub-element.
            e.currentBuffer.addElement(elem, elemType='px')
        if elem != 'x':
            # Dump the start element and its attributes. But as a preamble,
            # manage special attributes that could not be dumped at all, like
            # "selected" or "checked".
            hook = None
            ignorableAttrs = self.pxAttributes
            for name in self.noDumpTags:
                if (name in attrs) and attrs[name].startswith(':'):
                    hook = (name, attrs[name][1:])
                    ignorableAttrs += (name,)
                    break
            e.currentBuffer.dumpStartElement(elem, attrs,
                ignoreAttrs=ignorableAttrs, noEndTag=elem in SELF_CLOSING,
                hook=hook, renamedAttrs=self.renamedAttributes)

    def endElement(self, elem):
        e = self.env
        # Manage the potentially collected Python expression in
        # e.currentContent.
        if e.currentContent:
            e.currentBuffer.addExpression(e.currentContent)
            e.currentContent = ''
        # Dump the end element into the current buffer
        if (elem != 'x') and (elem not in SELF_CLOSING):
            e.currentBuffer.dumpEndElement(elem)
        # If this element is the main element of the current buffer, we must
        # pop it and continue to work in the parent buffer.
        if e.isActionElem(elem):
            # Is it the buffer main element?
            isMainElement = e.currentBuffer.isMainElement(elem)
            # Unreference the element among buffer.elements
            e.currentBuffer.unreferenceElement(elem)
            if isMainElement:
                # Continue to work in the parent buffer
                e.currentBuffer = e.currentBuffer.parent

    def characters(self, content):
        e = self.env
        if not e.currentContent and content.startswith(':'):
            # This content is not static content to dump as-is into the result:
            # it is a Python expression.
            e.currentContent += content[1:]
        elif e.currentContent:
            # We continue to dump the Python expression
            e.currentContent += content
        else:
            # Remove blanks which are there only for improving PX's readability
            if content == ' ':
                pass # Keep individual spaces
            elif content.isspace():
                return
            e.currentBuffer.dumpContent(content)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# -*- coding: utf-8 -*-
'''This folder contains copies of external, "authentic" data, stored as text
   files, like ISO 639.2 country codes. In this package, corresponding Python
   classes are available for accessing the data in the text files.'''

# ------------------------------------------------------------------------------
import os, os.path

# List of names of language in their own language ------------------------------
# It was copied from Plone 2.5.5 (PloneLanguageTool), don't know any "authentic
# source" for that.
nativeNames = {
  'aa' : 'магIарул мацI',
  'ab' : 'бызшәа',
  'af' : 'Afrikaans',
  'am' : 'አማርኛ',
  'ar' : 'العربية',
  'as' : 'অসমিয়া',
  'ay' : 'Aymara',
  'az' : 'Azəri Türkçəsi',
  'ba' : 'Bashkir',
  'be' : 'Беларускі',
  'bg' : 'Български',
  'bh' : 'Bihari',
  'bi' : 'Bislama',
  'bn' : 'বাংলা',
  'bo' : 'བོད་སྐད་',
  'bs' : 'Bosanski',
  'br' : 'Brezhoneg',
  'ca' : 'Català',
  'ch' : 'Chamoru',
  'co' : 'Corsu',
  'cs' : 'Čeština',
  'cy' : 'Cymraeg',
  'da' : 'Dansk',
  'de' : 'Deutsch',
  'dz' : 'རྫོང་ཁ',
  'el' : 'Ελληνικά',
  'en' : 'English',
  'eo' : 'Esperanto',
  'es' : 'Español',
  'et' : 'Eesti',
  'eu' : 'Euskara',
  'fa' : 'فارسی',
  'fi' : 'Suomi',
  'fj' : 'Fiji',
  'fo' : 'Føroyska',
  'fr' : 'Français',
  'fy' : 'Frysk',
  'ga' : 'Gaeilge',
  'gd' : 'Gàidhlig',
  'gl' : 'Galego',
  'gn' : 'Guarani',
  'gu' : 'ગુજરાતી',
  'gv' : 'Gaelg',
  'ha' : 'هَوُس',
  'he' : 'עברית',
  'hi' : 'हिंदी',
  'hr' : 'Hrvatski',
  'hu' : 'Magyar',
  'hy' : 'Հայերէն',
  'ia' : 'Interlingua',
  'id' : 'Bahasa Indonesia',
  'ie' : 'Interlingue',
  'ik' : 'Inupiak',
  'is' : 'Íslenska',
  'it' : 'Italiano',
  'iu' : 'ᐃᓄᒃᑎᑐᑦ',
  'ja' : '日本語',
  'jbo': 'lojban',
  'jw' : 'Basa Jawi',
  'ka' : 'ქართული',
  'kk' : 'ﻗﺎﺯﺍﻗﺸﺎ',
  'kl' : 'Greenlandic',
  'km' : 'ខ្មែរ',
  'kn' : 'ಕನ್ನಡ',
  'ko' : '한국어',
  'ks' : 'काऽशुर',
  'ku' : 'Kurdí',
  'kw' : 'Kernewek',
  'ky' : 'Кыргыз',
  'la' : 'Latin',
  'lb' : 'Lëtzebuergesch',
  'li' : 'Limburgs',
  'ln' : 'Lingala',
  'lo' : 'ພາສາລາວ',
  'lt' : 'Lietuviskai',
  'lv' : 'Latviešu',
  'mg' : 'Malagasy',
  'mi' : 'Maori',
  'mk' : 'Македонски',
  'ml' : 'മലയാളം',
  'mn' : 'Монгол',
  'mo' : 'Moldavian',
  'mr' : 'मराठी',
  'ms' : 'Bahasa Melayu',
  'mt' : 'Malti',
  'my' : 'Burmese',
  'na' : 'Nauru',
  'ne' : 'नेपाली',
  'nl' : 'Nederlands',
  'no' : 'Norsk',
  'nn' : 'Nynorsk',
  'oc' : 'Languedoc',
  'om' : 'Oromo',
  'or' : 'ଓଡ଼ିଆ',
  'pa' : 'ਪੰਜਾਬੀ',
  'pl' : 'Polski',
  'ps' : 'پښتو',
  'pt' : 'Português',
  'qu' : 'Quechua',
  'rm' : 'Rumantsch',
  'rn' : 'Kirundi',
  'ro' : 'Română',
  'ru' : 'Русский',
  'rw' : 'Kiyarwanda',
  'sa' : 'संस्कृत',
  'sd' : 'Sindhi',
  'se' : 'Northern Sámi',
  'sg' : 'Sangho',
  'sh' : 'Serbo-Croatian',
  'si' : 'Singhalese',
  'sk' : 'Slovenčina',
  'sl' : 'Slovenščina',
  'sm' : 'Samoan',
  'sn' : 'Shona',
  'so' : 'Somali',
  'sq' : 'Shqip',
  'sr' : 'српски',
  'ss' : 'Siswati',
  'st' : 'Sesotho',
  'su' : 'Sudanese',
  'sv' : 'Svenska',
  'sw' : 'Kiswahili',
  'ta' : 'தமிழ',
  'te' : 'తెలుగు',
  'tg' : 'Тоҷики',
  'th' : 'ไทย',
  'ti' : 'ትግርኛ',
  'tk' : 'түркmенче',
  'tl' : 'Tagalog',
  'tn' : 'Setswana',
  'to' : 'Lea faka-Tonga',
  'tr' : 'Türkçe',
  'ts' : 'Tsonga',
  'tt' : 'татарча',
  'tw' : 'Twi',
  'ug' : 'Uigur',
  'uk' : 'Українська',
  'ur' : 'اردو',
  'uz' : 'Ўзбекча',
  'vi' : 'Tiếng Việt',
  'vo' : 'Volapük',
  'wa' : 'Walon',
  'wo' : 'Wolof',
  'xh' : 'isiXhosa',
  'yi' : 'ײִדיש',
  'yo' : 'Yorùbá',
  'za' : 'Zhuang',
  'zh' : '中文',
  'zu' : 'isiZulu'
}
# List of languages having direction right-to-left (RTL) -----------------------
rtlLanguages = ('ar', 'he', 'fa')

# Countries of the "euro" zone
vatEuroCountries = ('AT', 'BE', 'BG', 'CY', 'CZ', 'DE', 'DK', 'EE', 'GR', 'ES',
                    'FI', 'FR', 'GB', 'HU', 'IE', 'IT', 'LT', 'LU', 'LV', 'MT',
                    'NL', 'PL' 'PT', 'RO', 'SE', 'SI', 'SK')

# ------------------------------------------------------------------------------
class Languages:
    '''This class gives access to the language codes and names as standardized
       by ISO-639. The file has been downloaded in July 2009 from
       http://www.loc.gov/standards/iso639-2/ascii_8bits.html (UTF-8 version)'''

    def __init__(self):
        self.fileName = os.path.dirname(__file__) + '/LanguageCodesIso639.2.txt'
        self.languageCodes = []
        # Names of languages in English
        self.languageNames = []
        # Names of languages in their language. It is not part of ISO 639.2 and
        # is taken from dict languageNames above.
        self.nativeNames = []
        self.parseFile()

    def parseFile(self):
        '''Parses the language codes and names in the ISO file and puts them in
           self.languageCodes, self.languageNames and self.nativeNames.'''
        f = file(self.fileName)
        for line in f:
            if line.strip():
                lineElems = line.split('|')
                if lineElems[2].strip():
                    # I take only those that have a 2-chars ISO-639-1 code.
                    self.languageCodes.append(lineElems[2])
                    self.languageNames.append(lineElems[3])
                    if lineElems[2] in nativeNames:
                        self.nativeNames.append(nativeNames[lineElems[2]])
                    else:
                        # Put the english name nevertheless.
                        self.nativeNames.append(lineElems[3])
        f.close()

    def exists(self, code):
        '''Is p_code a valid 2-digits language code?'''
        return code in self.languageCodes

    def get(self, code):
        '''Returns information about the language whose code is p_code.'''
        try:
            iCode = self.languageCodes.index(code)
            return self.languageCodes[iCode], self.languageNames[iCode], \
                   self.nativeNames[iCode]
        except ValueError:
            return None, None, None

    def __repr__(self):
        i = -1
        res = ''
        for languageCode in self.languageCodes:
            i += 1
            res += 'Language: ' + languageCode + ' - ' + self.languageNames[i]
            res += '\n'
        return res
# We instantiate here Languages because it is used by the appy.gen languages
# management.
languages = Languages()

# Country codes ISO 3166-1. ----------------------------------------------------
class Countries:
    '''This class gives access to the country codes and names as standardized by
       ISO 3166-1. The file has been downloaded in March 2011 from
       http://www.iso.org/iso/country_codes/iso_3166_code_lists.htm
       (first line has been removed).'''

    def __init__(self):
        # This file has been downloaded from
        # http://www.iso.org/iso/country_codes.htm and converted to utf-8.
        self.fileName = os.path.dirname(__file__) + '/CountryCodesIso3166.1.txt'
        self.countryCodes = []
        # Names of countries in English
        self.countryNames = []
        self.parseFile()

    def parseFile(self):
        f = file(self.fileName)
        for line in f:
            if line.strip():
                name, code = line.split(';')
                self.countryCodes.append(code.strip())
                self.countryNames.append(name.strip())
        f.close()

    def exists(self, code):
        '''Is p_code a valid 2-digits country code?'''
        return code in self.countryCodes
# We instantiate here Countries because it is used by appy.gen for some field
# validations.
countries = Countries()

# ------------------------------------------------------------------------------
class BelgianCities:
    '''This class contains data about Belgian cities (postal codes). It creates
       a dictionary whose keys are postal codes and whose values are city names.
       The corresponding Excel file was downloaded on 2009-10-26 from
       https://www.post.be/site/fr/sse/advertising/addressed/biblio.html,
       converted to CSV (field separator being ";" field content is surrrounded
       by double quotes).'''

    def __init__(self):
        self.fileName = os.path.dirname(__file__) + '/BelgianCommunes.txt'
        self.data = {}
        self.parseFile()

    def parseFile(self):
        f = file(self.fileName)
        for line in f:
            if line.strip():
                lineElems = line.split(';')
                self.data[int(lineElems[0].strip('"'))]= lineElems[1].strip('"')

    def exists(self, postalCode):
        '''Is postalCode a valid Belgian postal code?'''
        return self.data.has_key(postalCode)
# ------------------------------------------------------------------------------



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

'''Test module'''



'''Parser allowing to retrieve Python objects from data defined in tabled within
    a LbreOffice Writer document (ODT).'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# The ODT document must follow these conventions.
# - Each table must have a first row with only one cell: the table name.
# - The other rows must all have the same number of columns. This number must
#   be strictly greater than 1.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import os, os.path, re
from collections import UserList, UserDict

from appy.xml import Parser
from appy.utils.zip import unzip
from appy.utils import path as putils

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class ParserError(Exception): pass
class TypeError(Exception): pass

# ParserError-related constants  - - - - - - - - - - - - - - - - - - - - - - - -
BAD_PARENT_ROW = 'For table "%s", you specified "%s" as parent table, but ' \
  'you referred to row number "%s" within the parent. This value must be a ' \
  'positive integer or zero (we start counting rows at 0).'
PARENT_NOT_FOUND = 'I cannot find table "%s" that you defined as being ' \
  'parent of "%s".'
TABLE_KEY_ERROR = 'Within a row of table "%s", you mention a column named ' \
  '"%s" which does not exist neither in "%s" itself, neither in its parent ' \
  'row(s). '
PARENT_ROW_NOT_FOUND = 'You specified table "%s" as inheriting from table ' \
  '"%s", row "%d", but this row does not exist (table "%s" as a length = %d). '\
  ' Note that we start counting rows at 0.'
PARENT_COLUMN_NOT_FOUND = 'You specified table "%s" as inheriting from table ' \
  '"%s", column "%s", but this column does not exist in table "%s" or parents.'
PARENT_ROW_COL_NOT_FOUND = 'You specified table "%s" as inheriting from ' \
  'table "%s", column "%s", value "%s", but it does not correspond to any ' \
  'row in table "%s".'
NO_ROWS_IN_TABLE_YET = 'In first row of table "%s", you use value \' " \' ' \
  'for referencing the cell value in previous row, which does not exist.'
VALUE_ERROR = 'Value error for column "%s" of table "%s". %s'
TYPE_ERROR = 'Type error for column "%s" of table "%s". %s'

# TypeError-related constants  - - - - - - - - - - - - - - - - - - - - - - - - -
LIST_TYPE_ERROR = 'Maximum number of nested lists is 4.'
BASIC_TYPE_ERROR = 'Letter "%s" does not correspond to any valid type. ' \
  'Valid types are f (float), i (int), g (long) and b (bool).'
BASIC_VALUE_ERROR = 'Value "%s" can\'t be converted to type "%s".'
LIST_VALUE_ERROR = 'Value "%s" is malformed: within it, %s. You should check ' \
  'the use of separators ( , : ; - ) to obtain a schema conform to the type ' \
  '"%s".'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Type:
    basicTypes = {'f': float, 'i':int, 'b':bool}
    separators = ['-', ';', ',', ':']

    def __init__(self, typeDecl):
        self.basicType = None # The Python basic type
        self.listNumber = 0
        # If = 1 : it is a list. If = 2: it is a list of lists. If = 3...
        self.analyseTypeDecl(typeDecl)
        if self.listNumber > 4:
            raise TypeError(LIST_TYPE_ERROR)
        self.name = self.computeName()

    def analyseTypeDecl(self, typeDecl):
        for char in typeDecl:
            if char == 'l':
                self.listNumber += 1
            else:
                # Get the basic type
                if char not in Type.basicTypes:
                    raise TypeError(BASIC_TYPE_ERROR % char)
                self.basicType = Type.basicTypes[char]
                break
        if not self.basicType:
            self.basicType = str

    def convertBasicValue(self, value):
        try:
            return self.basicType(value.strip())
        except ValueError:
            raise TypeError(BASIC_VALUE_ERROR % (value,
                                                 self.basicType.__name__))

    def convertValue(self, value):
        '''Converts a p_value which is a string into a value conform
        to self.'''
        if self.listNumber == 0:
            r = self.convertBasicValue(value)
        else:
            # Get separators in their order of appearance
            separators = []
            for char in value:
                if (char in Type.separators) and (char not in separators):
                    separators.append(char)            
            # Remove surplus separators
            if len(separators) > self.listNumber:
                nbOfSurplusSeps = len(separators) - self.listNumber
                separators = separators[nbOfSurplusSeps:]
            # If not enough separators, create corresponding empty lists.
            r = None
            innerList = None
            rIsComplete = False
            if len(separators) < self.listNumber:
                if not value:
                    r = []
                    rIsComplete = True
                else:
                    # Begin with empty list(s)
                    nbOfMissingSeps = self.listNumber - len(separators)
                    r = []
                    innerList = r
                    for i in range(nbOfMissingSeps-1):
                        newInnerList = []
                        innerList.append(newInnerList)
                        innerList = newInnerList
            # We can now convert the value
            separators.reverse()
            if innerList is not None:
                innerList.append(self.convertListItem(value, separators))
            elif not rIsComplete:
                try:
                    r = self.convertListItem(value, separators)
                except TypeError as te:
                    raise TypeError(LIST_VALUE_ERROR % (value, te, self.name))
        return r

    def convertListItem(self, stringItem, remainingSeps):
        if not remainingSeps:
            r = self.convertBasicValue(stringItem)
        else:
            curSep = remainingSeps[0]
            tempRes = stringItem.split(curSep)
            if (len(tempRes) == 1) and (not tempRes[0]):
                # There was no value within value, so we produce an empty list.
                r = []
            else:
                r = []
                for tempItem in tempRes:
                    r.append(self.convertListItem(tempItem, remainingSeps[1:]))
        return r

    def computeName(self):
        prefix = 'list of ' * self.listNumber
        return '<%s%s>' % (prefix, self.basicType.__name__)

    def __repr__(self): return self.name

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class TableRow(UserDict):
    def __init__(self, table):
        UserDict.__init__(self)
        self.table = table

    def __getitem__(self, key):
        '''This method "implements" row inheritance: if the current row does
           not have an element with p_key, it looks in the parent row of this
           row, via the parent table self.table.'''
        # Return the value from this dict if present
        if key in self: return UserDict.__getitem__(self, key)
        # Try to get the value from the parent row
        keyError = False
        t = self.table
        # Get the parent row
        if t.parent:
            if isinstance(t.parentRow, int):
                if t.parentRow < len(t.parent):
                    try:
                        r = t.parent[t.parentRow][key]
                    except KeyError:
                        keyError = True
                else:
                    raise ParserError(PARENT_ROW_NOT_FOUND % (t.name,
                      t.parent.name, t.parentRow, t.parent.name, len(t.parent)))
            else:
                tColumn, tValue = t.parentRow
                # Get the 1st row having tColumn = tValue
                rowFound = False
                for row in t.parent:
                    try:
                        curVal = row[tColumn]
                    except KeyError:
                        raise ParserError(PARENT_COLUMN_NOT_FOUND % (t.name,
                                         t.parent.name, tColumn, t.parent.name))
                    if curVal == tValue:
                        rowFound = True
                        try:
                            r = row[key]
                        except KeyError:
                            keyError = True
                        break
                if not rowFound:
                    raise ParserError(PARENT_ROW_COL_NOT_FOUND % (t.name,
                                 t.parent.name, tColumn, tValue, t.parent.name))
        else:
            keyError = True
        if keyError:
            raise KeyError(TABLE_KEY_ERROR % (t.name, key, t.name))
        return r

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Table(UserList):
    nameRex = re.compile('([^\(]+)(?:\((.*)\))?')

    def __init__(self):
        UserList.__init__(self)
        self.name = ''
        # Column names
        self.columns = []
        # Column types. If no type is defined for some column, value "None" will
        # be stored and wille correspond to default type "str".
        self.columnTypes = []
        # The parent table
        self.parent = None
        # The parent row, within the parent table
        self.parentRow = None

    def setName(self):
        '''Parses the table name and extracts from it the potential reference
           to a parent table.'''
        name = self.name.strip()
        elems = Table.nameRex.search(name)
        name, parentSpec = elems.groups()
        self.name = name
        if parentSpec:
            r = parentSpec.split(':')
            if len(r) == 1:
                self.parent = parentSpec.strip()
                self.parentRow = 0
            else:
                self.parent = r[0].strip()
                r = r[1].split('=')
                if len(r) == 1:
                    try:
                        self.parentRow = int(r[0])
                    except ValueError:
                        msg = BAD_PARENT_ROW % (self.name, self.parent, r[0])
                        raise ParserError(msg)
                    if self.parentRow < 0:
                        msg = BAD_PARENT_ROW % (self.name, self.parent, r[0])
                        raise ParserError(msg)
                else:
                    self.parentRow = (r[0].strip(), r[1].strip())

    def addRow(self):
        '''Adds a row of data into this table, as a TableRow instance. This
           consists in converting the list of row data we have already added
           into the table into a TableRow instance.'''
        data = self[-1]
        row = TableRow(self)
        for i in range(len(self.columns)):
            column = self.columns[i]
            value = data[i].strip()
            if value == '"':
                # Check if a previous row exists
                if len(self) == 1:
                    raise ParserError(NO_ROWS_IN_TABLE_YET % self.name)
                value = self[-2][column]
            else:
                # Get the column type and convert the value to this type
                type = self.columnTypes[i]
                if type:
                    try:
                        value = type.convertValue(value)
                    except TypeError as te:
                        raise ParserError(VALUE_ERROR % (column, self.name, te))
            row[self.columns[i]] = value
        self[-1] = row

    def addColumn(self):
        '''A new column has been added into self.columns. Extract its type if
           defined.'''
        # Extract the parsed column name
        name = self.columns[-1].strip()
        type = None
        if ':' in name:
            # We have a type declaration
            name, typeDecl = name.split(':')
            try:
                type = Type(typeDecl.strip())
            except TypeError as te:
                raise ParserError(TYPE_ERROR % (name, self.name, te))
        # Update the lists of header names and types
        self.columns[-1] = name
        self.columnTypes.append(type)
        # Create a new empty column for the next one
        self.columns.append('')

    def dump(self, withContent=True):
        r = 'Table "%s"' % self.name
        if self.parent:
            r += ' extends table "%s"' % self.parent.name
            if isinstance(self.parentRow, int):
                r += '(%d)' % self.parentRow
            else:
                r += '(%s=%s)' % self.parentRow
        if withContent:
            r += '\n'
            for line in self:
                r += str(line)
        return r

    def instanceOf(self, tableName):
        r = False
        if self.parent:
            if self.parent.name == tableName:
                r = True
            else:
                r = self.parent.instanceOf(tableName)
        return r

    def asDict(self):
        '''If this table as only 2 columns named "key" and "value", it can be
           represented as a Python dict. This method produces this dict.'''
        infoDict = {}
        if self.parent:
            for info in self.parent:
              infoDict[info["key"]] = info["value"]
        for info in self:
            infoDict[info["key"]] = info["value"]
        return infoDict

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class OdtTablesParser(Parser):
    PARSING = 0
    PARSING_TABLE_NAME = 1
    PARSING_TABLE_HEADERS = 2
    PARSING_DATA_ROW = 3

    def startDocument(self):
        Parser.startDocument(self)
        # Where to store the parsed tables, keyed by their names
        self.r = {}
        # The currently walked table
        self.currentTable = None
        # The currently walked data row, as list of strings (cell content)
        self.currentRow = None
        # The initial parsing state
        self.state = self.PARSING

    def startElement(self, tag, attrs):
        if tag == 'table:table':
            # A new table is encountered
            self.currentTable = Table()
        elif tag == 'table:table-row':
            if not self.currentTable.name:
                # This is the first row (table name)
                self.state = self.PARSING_TABLE_NAME
            elif not self.currentTable.columns:
                # This is the second row (column headers)
                self.state = self.PARSING_TABLE_HEADERS
                self.currentTable.columns.append('')
            else:
                # This is a data row
                self.state = self.PARSING_DATA_ROW
                self.currentTable.append([''])
        elif tag == 'table:table-cell':
            pass

    def endElement(self, tag):
        if tag == 'table:table':
            # Store the completelty parsed table in self.res
            table = self.currentTable
            self.r[table.name] = table
            self.currentTable = None
        elif tag == 'table:table-row':
            if self.state == self.PARSING_TABLE_NAME:
                # We have finished to parse the first row (table name)
                self.currentTable.setName()
                self.state = self.PARSING
            elif self.state == self.PARSING_TABLE_HEADERS:
                # We have finished to parse the second row
                del self.currentTable.columns[-1]
                self.state = self.PARSING
            elif self.state == self.PARSING_DATA_ROW:
                # We have finished parsing a data row
                del self.currentTable[-1][-1]
                self.currentTable.addRow()
                self.state = self.PARSING
        elif tag == 'table:table-cell':
            if self.state == self.PARSING_TABLE_HEADERS:
                # We have finished parsing a header value. Add a new one.
                self.currentTable.addColumn()
            elif self.state == self.PARSING_DATA_ROW:
                # We have finished parsing a cell value. Add a new one.
                self.currentTable[-1].append('')

    def characters(self, content):
        # Get the table name
        if self.state == self.PARSING_TABLE_NAME:
            self.currentTable.name += content
        elif self.state == self.PARSING_TABLE_HEADERS:
            self.currentTable.columns[-1] += content
        elif self.state == self.PARSING_DATA_ROW:
            self.currentTable[-1][-1] += content

    def endDocument(self):
        Parser.endDocument(self)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class TablesParser:
    def __init__(self, path):
        # The path to the ODT file containing the tables
        self.path = path
        # Parsed tables will be stored in this dict
        self.tables = None

    def linkTables(self):
        '''Resolve parend/child links between parsed tables'''
        for name, table in self.tables.items():
            if not table.parent: continue
            if table.parent not in self.tables:
                raise ParserError(PARENT_NOT_FOUND % (table.parent, table.name))
            table.parent = self.tables[table.parent]

    def run(self):
        '''Unzip the ODT file and parse tables within content.xml'''
        # Create a folder in the OS temp folder
        folder = putils.getOsTempFolder(sub=True)
        # Unzip the file in the OS temp folder
        unzip(str(self.path), folder)
        # Parse content.xml
        contentXml = os.path.join(folder, 'content.xml')
        self.tables = OdtTablesParser().parse(contentXml, source='file')
        # Revolve parent/child table links
        self.linkTables()
        # Delete the folder
        putils.FolderDeleter.delete(folder)
        return self.tables
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



xhtmlInput = '''
<div class="document">
<p>Some <strong>bold</strong> and some <em>italic</em> text.</p>
<p>A new paragraph.</p>
<p>A list with three items:</p>
<ul> 
<li>the first item</li> 
<li>another item</li> 
<li>the last item</li> 
</ul> 
<p>A last paragraph.</p>
</div>
<table border="1" cellpadding="1" cellspacing="1" style="width:500px">
	<tbody>
		<tr>
			<td>A1</td>
			<td colspan="1" rowspan="2">B12</td>
			<td>C1</td>
		</tr>
		<tr>
			<td>A2</td>
			<td>C2</td>
		</tr>
		<tr>
			<td colspan="2" rowspan="1">A3A2</td>
			<td>A3</td>
		</tr>
	</tbody>
</table>
'''



'''This module is about converting chunks of (X)HTML code into ODF code that
   pod can then integrate into an pod template, via a statement like
   
                      "do ... from xhtml(...)"
'''
